import os
import shutil
import subprocess
import sys
from unittest import mock

import pytest
import tomli_w

from briefcase.console import Console, Log, LogLevel
from briefcase.exceptions import (
    BriefcaseCommandError,
    NetworkFailure,
    UnsupportedHostError,
)
from briefcase.integrations.docker import Docker, DockerAppContext
from briefcase.integrations.linuxdeploy import LinuxDeploy, LinuxDeployBase
from briefcase.platforms.linux.appimage import LinuxAppImageBuildCommand

from ....utils import create_file


@pytest.fixture
def first_app(first_app_config, tmp_path):
    """A fixture for the first app, rolled out on disk."""
    # Make it look like the template has been generated
    app_dir = tmp_path / "base_path/build/first-app/linux/appimage/First App.AppDir"
    (app_dir / "usr/app/support").mkdir(parents=True, exist_ok=True)
    (app_dir / "usr/app_packages/firstlib").mkdir(parents=True, exist_ok=True)
    (app_dir / "usr/app_packages/secondlib").mkdir(parents=True, exist_ok=True)

    # Create some .so files
    (app_dir / "usr/app/support/support.so").touch()
    (app_dir / "usr/app_packages/firstlib/first.so").touch()
    (app_dir / "usr/app_packages/secondlib/second_a.so").touch()
    (app_dir / "usr/app_packages/secondlib/second_b.so").touch()

    return first_app_config


@pytest.fixture
def build_command(tmp_path, first_app_config):
    command = LinuxAppImageBuildCommand(
        logger=Log(),
        console=Console(),
        base_path=tmp_path / "base_path",
        data_path=tmp_path / "briefcase",
        apps={"first": first_app_config},
    )
    command.tools.host_os = "Linux"
    command.tools.host_arch = "x86_64"
    command.use_docker = False
    command.extra_docker_build_args = []
    command._briefcase_toml[first_app_config] = {
        "paths": {
            "app_path": "First App.AppDir/usr/app",
            "app_packages_path": "First App.AppDir/usr/app_packages",
        }
    }

    # Reset `os` mock without `spec` so tests can run on Windows where os.getuid doesn't exist.
    command.tools.os = mock.MagicMock()
    # Mock user and group IDs for docker image
    command.tools.os.environ = mock.MagicMock()
    command.tools.os.environ.__getitem__.return_value = (
        "/usr/local/bin:/usr/bin:/path/to/somewhere"
    )
    command.tools.os.environ.copy.return_value = {
        "PATH": "/usr/local/bin:/usr/bin:/path/to/somewhere"
    }

    # mock user and group IDs for docker build
    command.tools.os.getuid.return_value = 1000
    command.tools.os.getgid.return_value = 1001

    # Store the underlying subprocess instance
    command._subprocess = mock.MagicMock(spec_set=subprocess)
    command.tools.subprocess._subprocess = command._subprocess

    # mock `echo $PATH` check_output call
    command.tools.subprocess._subprocess.check_output.return_value = (
        "/usr/local/bin:/usr/bin:/path/to/somewhere"
    )

    # Short circuit the process streamer
    wait_bar_streamer = mock.MagicMock()
    wait_bar_streamer.stdout.readline.return_value = ""
    wait_bar_streamer.poll.return_value = 0
    command.tools.subprocess._subprocess.Popen.return_value.__enter__.return_value = (
        wait_bar_streamer
    )

    command.tools.linuxdeploy = LinuxDeploy(command.tools)
    return command


def test_verify_tools_wrong_platform(build_command):
    """If we're not on Linux, the build fails."""

    build_command.tools.host_os = "TestOS"
    build_command.build_app = mock.MagicMock()
    build_command.tools.file.download = mock.MagicMock()

    # Try to invoke the build
    with pytest.raises(UnsupportedHostError):
        build_command()

    # The download was not attempted
    assert build_command.tools.file.download.call_count == 0

    # But it failed, so the file won't be made executable...
    assert build_command.tools.os.chmod.call_count == 0

    # and no build will be attempted
    assert build_command.build_app.call_count == 0


def test_verify_tools_download_failure(build_command):
    """If the build tools can't be retrieved, the build fails."""
    # Remove linuxdeploy tool so download is attempted
    delattr(build_command.tools, "linuxdeploy")

    build_command.build_app = mock.MagicMock()
    build_command.tools.file.download = mock.MagicMock(
        side_effect=NetworkFailure("mock")
    )

    # Try to invoke the build
    with pytest.raises(BriefcaseCommandError):
        build_command()

    # The download was attempted
    build_command.tools.file.download.assert_called_with(
        url="https://github.com/linuxdeploy/linuxdeploy/releases/download/continuous/linuxdeploy-x86_64.AppImage",
        download_path=build_command.tools.base_path,
        role="linuxdeploy",
    )

    # But it failed, so the file won't be made executable...
    assert build_command.tools.os.chmod.call_count == 0

    # and no build will be attempted
    assert build_command.build_app.call_count == 0


@pytest.mark.parametrize("debug_mode", (True, False))
def test_build_appimage(build_command, first_app, debug_mode, tmp_path, sub_stream_kw):
    """A Linux app can be packaged as an AppImage."""
    # Enable verbose tool logging
    if debug_mode:
        build_command.tools.logger.verbosity = LogLevel.DEEP_DEBUG

    build_command.verify_app_tools(first_app)
    build_command.build_app(first_app)

    # linuxdeploy was invoked
    app_dir = tmp_path / "base_path/build/first-app/linux/appimage/First App.AppDir"
    expected_env = {
        "PATH": "/usr/local/bin:/usr/bin:/path/to/somewhere",
        "LINUXDEPLOY_OUTPUT_VERSION": "0.0.1",
        "DISABLE_COPYRIGHT_FILES_DEPLOYMENT": "1",
        "APPIMAGE_EXTRACT_AND_RUN": "1",
        "ARCH": "x86_64",
    }
    if debug_mode:
        expected_env["DEBUG"] = "1"
    build_command._subprocess.Popen.assert_called_with(
        [
            os.fsdecode(tmp_path / "briefcase/tools/linuxdeploy-x86_64.AppImage"),
            "--appdir",
            os.fsdecode(app_dir),
            "--desktop-file",
            os.fsdecode(app_dir / "com.example.first-app.desktop"),
            "--output",
            "appimage",
            "-v0" if debug_mode else "-v1",
            "--deploy-deps-only",
            os.fsdecode(app_dir / "usr/app/support"),
            "--deploy-deps-only",
            os.fsdecode(app_dir / "usr/app_packages/firstlib"),
            "--deploy-deps-only",
            os.fsdecode(app_dir / "usr/app_packages/secondlib"),
        ],
        env=expected_env,
        cwd=os.fsdecode(tmp_path / "base_path/build/first-app/linux/appimage"),
        **sub_stream_kw,
    )
    # Binary is marked executable
    build_command.tools.os.chmod.assert_called_with(
        tmp_path
        / "base_path/build/first-app/linux/appimage/First_App-0.0.1-x86_64.AppImage",
        0o755,
    )


@pytest.mark.skipif(
    sys.platform == "win32",
    reason="Windows paths can't be passed to linuxdeploy",
)
def test_build_appimage_with_plugin(build_command, first_app, tmp_path, sub_stream_kw):
    """A Linux app can be packaged as an AppImage with a plugin."""
    # Mock the existence of some plugins
    gtk_plugin_path = (
        tmp_path / "briefcase/tools/linuxdeploy_plugins/gtk/linuxdeploy-plugin-gtk.sh"
    )
    gtk_plugin_path.parent.mkdir(parents=True)
    gtk_plugin_path.touch()

    local_file_plugin_path = tmp_path / "local/linuxdeploy-plugin-something.sh"
    local_file_plugin_path.parent.mkdir(parents=True)
    local_file_plugin_path.touch()

    # Configure the app to use the plugins
    first_app.linuxdeploy_plugins = [
        "DEPLOY_GTK_VERSION=3 gtk",
        os.fsdecode(local_file_plugin_path),
    ]

    # Build the app
    build_command.verify_app_tools(first_app)
    build_command.build_app(first_app)

    # linuxdeploy was invoked
    app_dir = tmp_path / "base_path/build/first-app/linux/appimage/First App.AppDir"
    build_command._subprocess.Popen.assert_called_with(
        [
            os.fsdecode(tmp_path / "briefcase/tools/linuxdeploy-x86_64.AppImage"),
            "--appdir",
            os.fsdecode(app_dir),
            "--desktop-file",
            os.fsdecode(app_dir / "com.example.first-app.desktop"),
            "--output",
            "appimage",
            "-v1",
            "--deploy-deps-only",
            os.fsdecode(app_dir / "usr/app/support"),
            "--deploy-deps-only",
            os.fsdecode(app_dir / "usr/app_packages/firstlib"),
            "--deploy-deps-only",
            os.fsdecode(app_dir / "usr/app_packages/secondlib"),
            "--plugin",
            "gtk",
            "--plugin",
            "something",
        ],
        env={
            "PATH": f"{gtk_plugin_path.parent}:{app_dir.parent}:/usr/local/bin:/usr/bin:/path/to/somewhere",
            "DEPLOY_GTK_VERSION": "3",
            "LINUXDEPLOY_OUTPUT_VERSION": "0.0.1",
            "DISABLE_COPYRIGHT_FILES_DEPLOYMENT": "1",
            "APPIMAGE_EXTRACT_AND_RUN": "1",
            "ARCH": "x86_64",
        },
        cwd=os.fsdecode(tmp_path / "base_path/build/first-app/linux/appimage"),
        **sub_stream_kw,
    )
    # Local plugin marked executable
    build_command.tools.os.chmod.assert_any_call(
        tmp_path
        / "base_path/build/first-app/linux/appimage/linuxdeploy-plugin-something.sh",
        0o755,
    )
    # Binary is marked executable
    build_command.tools.os.chmod.assert_called_with(
        tmp_path
        / "base_path/build/first-app/linux/appimage/First_App-0.0.1-x86_64.AppImage",
        0o755,
    )


def test_build_failure(build_command, first_app, tmp_path, sub_stream_kw):
    """If linuxdeploy fails, the build is stopped."""
    # Mock a failure in the build
    build_command._subprocess.Popen.side_effect = subprocess.CalledProcessError(
        cmd=["linuxdeploy-x86_64.AppImage", "..."], returncode=1
    )

    # Invoking the build will raise an error.
    build_command.verify_app_tools(first_app)
    with pytest.raises(BriefcaseCommandError):
        build_command.build_app(first_app)

    # linuxdeploy was invoked
    app_dir = tmp_path / "base_path/build/first-app/linux/appimage/First App.AppDir"
    build_command._subprocess.Popen.assert_called_with(
        [
            os.fsdecode(tmp_path / "briefcase/tools/linuxdeploy-x86_64.AppImage"),
            "--appdir",
            os.fsdecode(app_dir),
            "--desktop-file",
            os.fsdecode(app_dir / "com.example.first-app.desktop"),
            "--output",
            "appimage",
            "-v1",
            "--deploy-deps-only",
            os.fsdecode(app_dir / "usr/app/support"),
            "--deploy-deps-only",
            os.fsdecode(app_dir / "usr/app_packages/firstlib"),
            "--deploy-deps-only",
            os.fsdecode(app_dir / "usr/app_packages/secondlib"),
        ],
        env={
            "PATH": "/usr/local/bin:/usr/bin:/path/to/somewhere",
            "LINUXDEPLOY_OUTPUT_VERSION": "0.0.1",
            "DISABLE_COPYRIGHT_FILES_DEPLOYMENT": "1",
            "APPIMAGE_EXTRACT_AND_RUN": "1",
            "ARCH": "x86_64",
        },
        cwd=os.fsdecode(tmp_path / "base_path/build/first-app/linux/appimage"),
        **sub_stream_kw,
    )

    # chmod isn't invoked if the binary wasn't created.
    assert build_command.tools.os.chmod.call_count == 0


@pytest.mark.skipif(
    sys.platform == "win32", reason="Windows paths aren't converted in Docker context"
)
def test_build_appimage_in_docker(
    build_command,
    first_app,
    sub_stream_kw,
    tmp_path,
    monkeypatch,
):
    """A Linux app can be packaged as an AppImage in a docker container."""

    # Enable docker, and move to a non-Linux OS.
    build_command.tools.host_os = "TestOS"
    build_command.use_docker = True

    # Provide Docker
    monkeypatch.setattr(
        Docker, "_is_user_mapping_enabled", mock.MagicMock(return_value=True)
    )
    build_command.tools.docker = Docker(tools=build_command.tools)
    # Provide Docker app context
    build_command.tools[first_app].app_context = DockerAppContext(
        tools=build_command.tools,
        app=first_app,
    )
    build_command.tools[first_app].app_context.prepare(
        image_tag=f"briefcase/com.example.first-app:py3.{sys.version_info.minor}",
        dockerfile_path=tmp_path
        / "base_path/build/first-app/linux/appimage/Dockerfile",
        app_base_path=tmp_path / "base_path",
        host_bundle_path=tmp_path / "base_path/build/first-app/linux/appimage",
        host_data_path=tmp_path / "briefcase",
        python_version=f"3.{sys.version_info.minor}",
    )

    build_command.build_app(first_app)

    # linuxdeploy was invoked inside Docker
    build_command._subprocess.Popen.assert_called_with(
        [
            "docker",
            "run",
            "--rm",
            "--volume",
            f"{tmp_path / 'base_path/build/first-app/linux/appimage'}:/app:z",
            "--volume",
            f"{build_command.data_path}:/briefcase:z",
            "--env",
            "LINUXDEPLOY_OUTPUT_VERSION=0.0.1",
            "--env",
            "DISABLE_COPYRIGHT_FILES_DEPLOYMENT=1",
            "--env",
            "APPIMAGE_EXTRACT_AND_RUN=1",
            "--env",
            "ARCH=x86_64",
            "--workdir",
            "/app",
            f"briefcase/com.example.first-app:py3.{sys.version_info.minor}",
            "/briefcase/tools/linuxdeploy-x86_64.AppImage",
            "--appdir",
            "/app/First App.AppDir",
            "--desktop-file",
            "/app/First App.AppDir/com.example.first-app.desktop",
            "--output",
            "appimage",
            "-v1",
            "--deploy-deps-only",
            "/app/First App.AppDir/usr/app/support",
            "--deploy-deps-only",
            "/app/First App.AppDir/usr/app_packages/firstlib",
            "--deploy-deps-only",
            "/app/First App.AppDir/usr/app_packages/secondlib",
        ],
        env={
            "PATH": "/usr/local/bin:/usr/bin:/path/to/somewhere",
            "DOCKER_CLI_HINTS": "false",
        },
        **sub_stream_kw,
    )
    # Binary is marked executable
    build_command.tools.os.chmod.assert_called_with(
        tmp_path
        / "base_path/build/first-app/linux/appimage/First_App-0.0.1-x86_64.AppImage",
        0o755,
    )


@pytest.mark.skipif(
    sys.platform == "win32", reason="Windows paths aren't converted in Docker context"
)
def test_build_appimage_with_plugins_in_docker(
    build_command,
    first_app,
    tmp_path,
    sub_stream_kw,
    monkeypatch,
):
    """A Linux app can be packaged as an AppImage with plugins in a Docker container."""
    # Mock the existence of some plugins
    gtk_plugin_path = (
        tmp_path / "briefcase/tools/linuxdeploy_plugins/gtk/linuxdeploy-plugin-gtk.sh"
    )
    gtk_plugin_path.parent.mkdir(parents=True)
    gtk_plugin_path.touch()

    local_file_plugin_path = tmp_path / "local/linuxdeploy-plugin-something.sh"
    local_file_plugin_path.parent.mkdir(parents=True)
    local_file_plugin_path.touch()

    # Configure the app to use the plugins
    first_app.linuxdeploy_plugins = [
        "DEPLOY_GTK_VERSION=3 gtk",
        os.fsdecode(local_file_plugin_path),
    ]

    build_command._subprocess.check_output.return_value = "/docker/bin:/docker/sbin"

    # Enable docker, and move to a non-Linux OS.
    build_command.tools.host_os = "TestOS"
    monkeypatch.setattr(LinuxDeployBase, "supported_host_os", {"TestOS"})
    build_command.use_docker = True

    # Provide Docker
    monkeypatch.setattr(
        Docker, "_is_user_mapping_enabled", mock.MagicMock(return_value=True)
    )
    build_command.tools.docker = Docker(tools=build_command.tools)
    # Provide Docker app context
    build_command.tools[first_app].app_context = DockerAppContext(
        tools=build_command.tools,
        app=first_app,
    )
    build_command.tools[first_app].app_context.prepare(
        image_tag=f"briefcase/com.example.first-app:py3.{sys.version_info.minor}",
        dockerfile_path=tmp_path
        / "base_path/build/first-app/linux/appimage/Dockerfile",
        app_base_path=tmp_path / "base_path",
        host_bundle_path=tmp_path / "base_path/build/first-app/linux/appimage",
        host_data_path=tmp_path / "briefcase",
        python_version=f"3.{sys.version_info.minor}",
    )

    build_command.build_app(first_app)

    # linuxdeploy was invoked inside Docker
    build_command._subprocess.Popen.assert_called_with(
        [
            "docker",
            "run",
            "--rm",
            "--volume",
            f"{tmp_path / 'base_path/build/first-app/linux/appimage'}:/app:z",
            "--volume",
            f"{build_command.data_path}:/briefcase:z",
            "--env",
            "DEPLOY_GTK_VERSION=3",
            "--env",
            (
                "PATH=/briefcase/tools/linuxdeploy_plugins/gtk"
                ":/app:/docker/bin:/docker/sbin"
            ),
            "--env",
            "LINUXDEPLOY_OUTPUT_VERSION=0.0.1",
            "--env",
            "DISABLE_COPYRIGHT_FILES_DEPLOYMENT=1",
            "--env",
            "APPIMAGE_EXTRACT_AND_RUN=1",
            "--env",
            "ARCH=x86_64",
            "--workdir",
            "/app",
            f"briefcase/com.example.first-app:py3.{sys.version_info.minor}",
            "/briefcase/tools/linuxdeploy-x86_64.AppImage",
            "--appdir",
            "/app/First App.AppDir",
            "--desktop-file",
            "/app/First App.AppDir/com.example.first-app.desktop",
            "--output",
            "appimage",
            "-v1",
            "--deploy-deps-only",
            "/app/First App.AppDir/usr/app/support",
            "--deploy-deps-only",
            "/app/First App.AppDir/usr/app_packages/firstlib",
            "--deploy-deps-only",
            "/app/First App.AppDir/usr/app_packages/secondlib",
            "--plugin",
            "gtk",
            "--plugin",
            "something",
        ],
        env={
            "PATH": "/usr/local/bin:/usr/bin:/path/to/somewhere",
            "DOCKER_CLI_HINTS": "false",
        },
        **sub_stream_kw,
    )
    # Local plugin marked executable
    build_command.tools.os.chmod.assert_any_call(
        tmp_path
        / "base_path/build/first-app/linux/appimage/linuxdeploy-plugin-something.sh",
        0o755,
    )
    # Binary is marked executable
    build_command.tools.os.chmod.assert_called_with(
        tmp_path
        / "base_path/build/first-app/linux/appimage/First_App-0.0.1-x86_64.AppImage",
        0o755,
    )


def test_build_appimage_with_support_package_update(
    build_command,
    first_app,
    sub_stream_kw,
    tmp_path,
    capsys,
):
    """If a support package update is performed, the user is warned."""
    # To trigger the app package update logic, we need to invoke the full build
    # command, and fake being on a verified Linux install with a generated app.
    build_command.tools.host_os = "Linux"

    # Mock shutil so we can check for folder deletion
    build_command.tools.shutil = mock.MagicMock(spec_set=shutil)

    # Mock downloads so we don't hit the network
    build_command.tools.file.download = mock.MagicMock()

    # Hard code a support revision so that the download support package is fixed,
    # and no linuxdeploy plugins.
    first_app.support_revision = "3.37.42+12345"
    first_app.linuxdeploy_plugins = []

    # Fake the existence of some source files.
    create_file(tmp_path / "base_path/src/first_app/app.py", "print('an app')")

    # Mock the generated app template
    (build_command.bundle_path(first_app) / "src").mkdir(parents=True)

    # Populate a briefcase.toml that mirrors a real Windows app
    with (build_command.bundle_path(first_app) / "briefcase.toml").open("wb") as f:
        index = {
            "paths": {
                "app_path": "First App.AppDir/usr/app",
                "app_package_path": "First App.AppDir/usr/app_packages",
                "support_path": "First App.AppDir/usr",
            }
        }
        tomli_w.dump(index, f)

    # Build the app with a support package update
    build_command(first_app, update_support=True)

    # linuxdeploy was invoked
    app_dir = tmp_path / "base_path/build/first-app/linux/appimage/First App.AppDir"
    build_command._subprocess.Popen.assert_called_with(
        [
            os.fsdecode(tmp_path / "briefcase/tools/linuxdeploy-x86_64.AppImage"),
            "--appdir",
            os.fsdecode(app_dir),
            "--desktop-file",
            os.fsdecode(app_dir / "com.example.first-app.desktop"),
            "--output",
            "appimage",
            "-v1",
            "--deploy-deps-only",
            os.fsdecode(app_dir / "usr/app/support"),
            "--deploy-deps-only",
            os.fsdecode(app_dir / "usr/app_packages/firstlib"),
            "--deploy-deps-only",
            os.fsdecode(app_dir / "usr/app_packages/secondlib"),
        ],
        env={
            "PATH": "/usr/local/bin:/usr/bin:/path/to/somewhere",
            "LINUXDEPLOY_OUTPUT_VERSION": "0.0.1",
            "DISABLE_COPYRIGHT_FILES_DEPLOYMENT": "1",
            "APPIMAGE_EXTRACT_AND_RUN": "1",
            "ARCH": "x86_64",
        },
        cwd=os.fsdecode(tmp_path / "base_path/build/first-app/linux/appimage"),
        **sub_stream_kw,
    )
    # Binary is marked executable
    build_command.tools.os.chmod.assert_called_with(
        tmp_path
        / "base_path/build/first-app/linux/appimage/First_App-0.0.1-x86_64.AppImage",
        0o755,
    )



from __future__ import annotations

import argparse
import sys
from argparse import RawDescriptionHelpFormatter

from briefcase import __version__
from briefcase.commands import (
    BuildCommand,
    ConvertCommand,
    CreateCommand,
    DevCommand,
    NewCommand,
    OpenCommand,
    PackageCommand,
    PublishCommand,
    RunCommand,
    UpdateCommand,
    UpgradeCommand,
)
from briefcase.commands.base import split_passthrough
from briefcase.console import MAX_TEXT_WIDTH, Console
from briefcase.platforms import get_output_formats, get_platforms

from .exceptions import InvalidFormatError, NoCommandError, UnsupportedCommandError

COMMANDS = [
    NewCommand,
    DevCommand,
    ConvertCommand,
    CreateCommand,
    OpenCommand,
    BuildCommand,
    UpdateCommand,
    RunCommand,
    PackageCommand,
    PublishCommand,
    UpgradeCommand,
]


def parse_cmdline(args, console: Console | None = None):
    """Parses the command line to determine the Command and its arguments.

    :param args: the arguments provided at the command line
    :param console: interface for interacting with the console
    :return: Command and command-specific arguments
    """
    if console is None:
        console = Console()

    platforms = get_platforms()

    briefcase_description = (
        "Briefcase is a tool for converting a Python project "
        "into a standalone native application for distribution."
    )

    description_max_pad_len = max(len(cmd.command) for cmd in COMMANDS) + 2
    command_description_list = "\n".join(
        f"  {cmd.command}{' ' * (description_max_pad_len - len(cmd.command))}{cmd.description}"
        for cmd in COMMANDS
    )

    platform_list = ", ".join(sorted(platforms, key=str.lower))

    additional_instruction = (
        "Each command, platform, and format has additional options. "
        "Use the -h option on a specific command for more details."
    )

    parser = argparse.ArgumentParser(
        prog="briefcase",
        description=console.textwrap(
            f"{briefcase_description}\n"
            "\n"
            "Commands:\n"
            f"{command_description_list}\n"
            "\n"
            "Platforms:\n"
            f"  {platform_list}\n"
            "\n"
            f"{additional_instruction}"
        ),
        usage="briefcase [-h] <command> [<platform>] [<format>] ...",
        add_help=False,
        formatter_class=(
            lambda prog: RawDescriptionHelpFormatter(prog, width=MAX_TEXT_WIDTH)
        ),
    )
    parser.add_argument("-V", "--version", action="version", version=__version__)

    # <command> isn't actually optional; but if it's marked as required,
    # there's no way to get help for subcommands. So; treat <command>
    # as optional, handle the case where <command> isn't provided
    # as the case where top-level help is displayed, and provide an explicit
    # usage string so that the instructions displayed are correct
    parser.add_argument(
        "command",
        choices=list(cmd.command for cmd in COMMANDS),
        metavar="command",
        nargs="?",
        help=argparse.SUPPRESS,
    )

    # To make the UX a little forgiving, we normalize *any* case to the case
    # actually used to register the platform. This function maps the lower-case
    # version of the registered name to the actual registered name.
    def normalize(name):
        return {n.lower(): n for n in platforms.keys()}.get(name.lower(), name)

    # argparse handles `--` specially, so make the passthrough args bypass the parser.
    def parse_known_args(args):
        args, passthrough = split_passthrough(args)
        options, extra = parser.parse_known_args(args)
        if passthrough:
            extra += ["--"] + passthrough
        return options, extra

    # Use parse_known_args to ensure any extra arguments can be ignored,
    # and parsed as part of subcommand handling. This will capture the
    # command, platform (filling a default if unspecified) and format
    # (with no value if unspecified).
    options, extra = parse_known_args(args)

    # If no command has been provided, display top-level help.
    if options.command is None:
        raise NoCommandError(parser.format_help())

    # Commands agnostic to the platform and format
    if options.command == "convert":
        Command = ConvertCommand
    elif options.command == "new":
        Command = NewCommand
    elif options.command == "dev":
        Command = DevCommand
    elif options.command == "upgrade":
        Command = UpgradeCommand

    # Commands dependent on the platform and format
    else:
        parser.add_argument(
            "platform",
            choices=list(platforms.keys()),
            default={
                "darwin": "macOS",
                "linux": "linux",
                "win32": "windows",
            }[sys.platform],
            metavar="platform",
            nargs="?",
            type=normalize,
            help="The platform to target (one of %(choices)s; default: %(default)s",
        )

        # <format> is also optional, with the default being platform dependent.
        # There's no way to encode option-dependent choices, so allow *any*
        # input, and we'll manually validate.
        parser.add_argument(
            "output_format",
            metavar="format",
            nargs="?",
            help="The output format to use (the available output formats are platform dependent)",
        )

        # Re-parse the arguments, now that we know it is a command that makes use
        # of platform/output_format.
        options, extra = parse_known_args(args)

        # Import the platform module
        platform_module = platforms[options.platform]

        # If the output format wasn't explicitly specified, check to see
        # Otherwise, extract and use the default output_format for the platform.
        if options.output_format is None:
            output_format = platform_module.DEFAULT_OUTPUT_FORMAT
        else:
            output_format = options.output_format

        output_formats = get_output_formats(options.platform)

        # Normalise casing of output_format to be more forgiving.
        output_format = {n.lower(): n for n in output_formats}.get(
            output_format.lower(), output_format
        )

        # We now know the command, platform, and format.
        # Get the command class that corresponds to that definition.
        try:
            format_module = output_formats[output_format]
            Command = getattr(format_module, options.command)
        except KeyError:
            raise InvalidFormatError(
                requested=output_format,
                choices=list(output_formats.keys()),
            )
        except AttributeError:
            raise UnsupportedCommandError(
                platform=options.platform,
                output_format=output_format,
                command=options.command,
            )

    return Command, extra



import pytest

from briefcase.console import LogLevel


def test_parse_options_no_overrides(base_command):
    """Command line options are parsed if provided without overrides."""
    options, overrides = base_command.parse_options(
        extra=(
            "-x",
            "wibble",
            "-r",
            "important",
        )
    )

    assert options == {
        "extra": "wibble",
        "mystery": None,
        "required": "important",
    }
    assert overrides == {}
    assert base_command.input.enabled
    assert base_command.logger.verbosity == LogLevel.INFO


def test_parse_options_with_overrides(base_command):
    """Command line options and overrides are parsed if provided."""
    options, overrides = base_command.parse_options(
        extra=(
            "-x",
            "wibble",
            "-r",
            "important",
            "-C",
            "width=10",
            "-C",
            "height=20",
        )
    )

    assert options == {
        "extra": "wibble",
        "mystery": None,
        "required": "important",
    }
    assert overrides == {
        "width": 10,
        "height": 20,
    }
    assert base_command.input.enabled
    assert base_command.logger.verbosity == LogLevel.INFO


@pytest.mark.parametrize(
    "verbosity, log_level",
    [
        ("", LogLevel.INFO),
        ("-v", LogLevel.VERBOSE),
        ("-vv", LogLevel.DEBUG),
        ("-vvv", LogLevel.DEEP_DEBUG),
        ("-vvvv", LogLevel.DEEP_DEBUG),
        ("-vvvvv", LogLevel.DEEP_DEBUG),
    ],
)
def test_verbosity(base_command, verbosity, log_level):
    """The logging level is set correctly for the verbosity."""
    base_command.parse_options(extra=filter(None, ("-r", "default", verbosity)))

    assert base_command.logger.verbosity == log_level


def test_missing_option(base_command, capsys):
    """If a required option isn't provided, an error is raised."""
    with pytest.raises(SystemExit) as excinfo:
        base_command.parse_options(extra=("-x", "wibble"))

    # Error code for a missing required option
    assert excinfo.value.code == 2
    # Error message about missing option is displayed
    err = capsys.readouterr().err
    assert "the following arguments are required: -r/--required" in err


def test_unknown_option(other_command, capsys):
    """If an unknown command is provided, it raises an error."""
    with pytest.raises(SystemExit) as excinfo:
        other_command.parse_options(extra=("-y", "because"))

    # Error code for an unknown option
    assert excinfo.value.code == 2
    # Error message about unknown option is displayed
    err = capsys.readouterr().err
    assert "unrecognized arguments: -y because" in err


def test_no_options(other_command, capsys):
    """If a command doesn't define options, any option is an error."""
    with pytest.raises(SystemExit) as excinfo:
        other_command.parse_options(extra=("-x", "wibble"))

    # Error code for an unknown option
    assert excinfo.value.code == 2
    # Error message about unknown option is displayed
    err = capsys.readouterr().err
    assert "unrecognized arguments: -x wibble" in err



class BriefcaseError(Exception):
    def __init__(self, error_code, skip_logfile=False):
        self.error_code = error_code
        self.skip_logfile = skip_logfile


class BriefcaseWarning(Exception):
    def __init__(self, error_code, msg):
        self.error_code = error_code
        self.msg = msg

    def __str__(self):
        return self.msg


class HelpText(BriefcaseError):
    """Exceptions that contain help text and shouldn't be displayed as an error."""


class NoCommandError(HelpText):
    def __init__(self, msg):
        super().__init__(error_code=-10, skip_logfile=True)
        self.msg = msg

    def __str__(self):
        return self.msg


class InvalidFormatError(BriefcaseError):
    def __init__(self, requested, choices):
        super().__init__(error_code=-21, skip_logfile=True)
        self.requested = requested
        self.choices = choices

    def __str__(self):
        choices = ", ".join(sorted(self.choices, key=str.lower))
        return f"Invalid format {self.requested!r}; (choose from: {choices})"


class UnsupportedCommandError(BriefcaseError):
    def __init__(self, platform, output_format, command):
        super().__init__(error_code=-30, skip_logfile=True)
        self.platform = platform
        self.output_format = output_format
        self.command = command

    def __str__(self):
        return (
            f"The {self.command} command for the {self.platform} {self.output_format} format "
            "has not been implemented (yet!)."
        )


class BriefcaseConfigError(BriefcaseError):
    def __init__(self, msg):
        super().__init__(error_code=100, skip_logfile=True)
        self.msg = msg

    def __str__(self):
        return f"Briefcase configuration error: {self.msg}"


class UnsupportedHostError(BriefcaseError):
    def __init__(self, reason):
        super().__init__(error_code=110, skip_logfile=True)
        self.msg = reason

    def __str__(self):
        return self.msg


class BriefcaseCommandError(BriefcaseError):
    def __init__(self, msg, skip_logfile=False):
        super().__init__(error_code=200, skip_logfile=skip_logfile)
        self.msg = msg

    def __str__(self):
        return self.msg


class NetworkFailure(BriefcaseCommandError):
    def __init__(self, action):
        self.action = action
        super().__init__(msg=f"Unable to {action}; is your computer offline?")


class MissingNetworkResourceError(BriefcaseCommandError):
    def __init__(self, url):
        self.url = url
        super().__init__(msg=f"Unable to download {url}; is the URL correct?")


class BadNetworkResourceError(BriefcaseCommandError):
    def __init__(self, url, status_code):
        self.url = url
        self.status_code = status_code
        super().__init__(msg=f"Unable to download {url} (status code {status_code})")


class MissingToolError(BriefcaseCommandError):
    def __init__(self, tool):
        self.tool = tool
        super().__init__(msg=f"Unable to locate {tool!r}. Has it been installed?")


class IncompatibleToolError(BriefcaseCommandError):
    def __init__(self, tool: str, env_var: str):
        self.tool = tool
        super().__init__(
            msg=f"""\
Briefcase cannot install {tool} on this machine.

Install {tool} manually and specify the installation directory in the {env_var} environment variable.
            """
        )


class NonManagedToolError(BriefcaseCommandError):
    def __init__(self, tool):
        self.tool = tool
        super().__init__(msg=f"{tool!r} is using an install that is user managed.")


class UpgradeToolError(BriefcaseCommandError):
    def __init__(self, error_msg):
        super().__init__(msg=error_msg, skip_logfile=True)


class TemplateUnsupportedVersion(BriefcaseCommandError):
    def __init__(self, briefcase_version):
        self.briefcase_version = briefcase_version
        super().__init__(
            f"Could not find a template branch for Briefcase {briefcase_version}."
        )


class InvalidTemplateRepository(BriefcaseCommandError):
    def __init__(self, template):
        self.template = template
        super().__init__(
            f"Unable to clone application template; is the template path {template!r} correct?"
        )


class UnsupportedPlatform(BriefcaseCommandError):
    def __init__(self, platform):
        self.platform = platform
        super().__init__(
            f"""\
App cannot be deployed on {platform}. This is probably because one or more
requirements (e.g., the GUI library) doesn't support {platform}.
"""
        )


class InvalidSupportPackage(BriefcaseCommandError):
    def __init__(self, filename):
        self.filename = filename
        super().__init__(f"Unable to unpack support package {str(filename)!r}.")


class InvalidStubBinary(BriefcaseCommandError):
    def __init__(self, filename):
        self.filename = filename
        super().__init__(f"Unable to unpack or copy stub binary {str(filename)!r}.")


class MissingAppMetadata(BriefcaseCommandError):
    def __init__(self, app_bundle_path):
        super().__init__(f"Unable to find {str(app_bundle_path / 'briefcase.toml')!r}")


class MissingSupportPackage(BriefcaseCommandError):
    def __init__(self, python_version_tag, platform, host_arch, is_32bit):
        self.python_version_tag = python_version_tag
        self.platform = platform
        self.host_arch = host_arch
        self.is_32bit = is_32bit
        platform_name = f"{'32 bit ' if is_32bit else ''}{platform}"
        super().__init__(
            f"""\
Unable to download {platform_name} support package for Python {self.python_version_tag} on {self.host_arch}.

This is likely because either Python {self.python_version_tag} and/or {self.host_arch} is not yet
supported on {platform_name}. You will need to:
    * Use an older version of Python; or
    * Compile your own custom support package.
"""
        )


class MissingStubBinary(BriefcaseCommandError):
    def __init__(self, python_version_tag, platform, host_arch, is_32bit):
        self.python_version_tag = python_version_tag
        self.platform = platform
        self.host_arch = host_arch
        self.is_32bit = is_32bit
        platform_name = f"{'32 bit ' if is_32bit else ''}{platform}"
        super().__init__(
            f"""\
Unable to download {platform_name} stub binary for Python {self.python_version_tag} on {self.host_arch}.

This is likely because either Python {self.python_version_tag} and/or {self.host_arch} is not yet
supported on {platform_name}. You will need to:
    * Use an older version of Python; or
    * Compile your own stub binary.
"""
        )


class RequirementsInstallError(BriefcaseCommandError):
    def __init__(self, install_hint=""):
        super().__init__(
            f"""\
Unable to install requirements. This may be because one of your
requirements is invalid, or because pip was unable to connect
to the PyPI server.{install_hint}
"""
        )


class MissingAppSources(BriefcaseCommandError):
    def __init__(self, src):
        self.src = src
        super().__init__(f"Application source {src!r} does not exist.")


class InvalidDeviceError(BriefcaseCommandError):
    def __init__(self, id_type, device):
        self.id_type = id_type
        self.device = device
        super().__init__(msg=f"Invalid device {id_type} {device!r}")


class CorruptToolError(BriefcaseCommandError):
    def __init__(self, tool):
        self.tool = tool
        super().__init__(msg=f"{tool!r} found, but it appears to be corrupted.")


class CommandOutputParseError(BriefcaseCommandError):
    def __init__(self, parse_error):
        super().__init__(msg=f"Unable to parse command output: {parse_error}")


class BriefcaseTestSuiteFailure(BriefcaseError):
    def __init__(self):
        super().__init__(error_code=1000, skip_logfile=True)


class NoDistributionArtefact(BriefcaseWarning):
    def __init__(self, msg):
        super().__init__(error_code=0, msg=msg)


class ParseError(Exception):
    """Raised by parser functions to signal parsing was unsuccessful."""



import subprocess
from collections import namedtuple
from pathlib import Path
from unittest.mock import MagicMock, call

import pytest

from briefcase.exceptions import BriefcaseCommandError, UnsupportedHostError
from briefcase.integrations.base import ToolCache
from briefcase.integrations.docker import Docker
from briefcase.integrations.subprocess import Subprocess

DOCKER_VERSION_MIN = "20.10"
VALID_DOCKER_VERSION = f"Docker version {DOCKER_VERSION_MIN}, build afacb8b\n"
VALID_DOCKER_INFO = "docker info printout"
VALID_BUILDX_VERSION = "github.com/docker/buildx v0.10.2 00ed17d\n"
VALID_USER_MAPPING_IMAGE_CACHE = "1ed313b0551f"
DOCKER_VERIFICATION_CALLS = [
    call(["docker", "--version"], env={"DOCKER_CLI_HINTS": "false"}),
    call(["docker", "info"], env={"DOCKER_CLI_HINTS": "false"}),
    call(["docker", "buildx", "version"], env={"DOCKER_CLI_HINTS": "false"}),
]


@pytest.fixture
def mock_tools(mock_tools) -> ToolCache:
    mock_tools.subprocess = MagicMock(spec_set=Subprocess)
    return mock_tools


@pytest.fixture
def mock_write_test_path(tmp_path, monkeypatch):
    """Mock the container write test path in to pytest's tmp directory."""
    write_test_path = tmp_path / "mock_write_test"
    # Wrap the path so read-only methods can be replaced
    write_test_path = MagicMock(wraps=write_test_path)
    monkeypatch.setattr(Docker, "_write_test_path", lambda self: write_test_path)
    return write_test_path


def test_short_circuit(mock_tools):
    """Tool is not created if already cached."""
    mock_tools.docker = "tool"

    tool = Docker.verify(mock_tools)

    assert tool == "tool"
    assert tool == mock_tools.docker


def test_unsupported_os(mock_tools):
    """When host OS is not supported, an error is raised."""
    mock_tools.host_os = "wonky"

    with pytest.raises(
        UnsupportedHostError,
        match=f"{Docker.name} is not supported on wonky",
    ):
        Docker.verify(mock_tools)


@pytest.mark.parametrize("host_os", ["Windows", "Linux", "Darwin"])
def test_docker_install_url(host_os):
    """Docker details available for each OS."""
    assert host_os in Docker.DOCKER_INSTALL_URL


@pytest.mark.parametrize(
    "version_output",
    [
        VALID_DOCKER_VERSION,
        # Docker version format on Ubuntu
        "Docker version 26.1.1, build 4cf5afa",
        # Docker version format on OpenSUSE Tumbleweed
        "Docker version 26.1.0-ce, build c8af8ebe4a89",
    ],
)
def test_docker_exists(mock_tools, user_mapping_run_calls, version_output, capsys):
    """If docker exists, the Docker wrapper is returned."""
    # Mock the return value of Docker Version
    mock_tools.subprocess.check_output.side_effect = [
        version_output,
        VALID_DOCKER_INFO,
        VALID_BUILDX_VERSION,
        VALID_USER_MAPPING_IMAGE_CACHE,
        "",  # touch host_write_file
        "",  # rm -f host_write_file
    ]

    # Invoke docker verify
    result = Docker.verify(mock_tools)

    # The verify call should return the Docker wrapper
    assert isinstance(result, Docker)

    # Docker version and plugins were verified and user mapping inspection occurred
    mock_tools.subprocess.check_output.assert_has_calls(
        DOCKER_VERIFICATION_CALLS + user_mapping_run_calls
    )

    # No console output
    output = capsys.readouterr()
    assert output.out == ""
    assert output.err == ""


def test_docker_doesnt_exist(mock_tools):
    """If docker doesn't exist, an error is raised."""
    # Mock Docker not installed on system
    mock_tools.subprocess.check_output.side_effect = FileNotFoundError

    # Invoke Docker verify
    with pytest.raises(BriefcaseCommandError):
        Docker.verify(mock_tools)

    # But docker was invoked
    mock_tools.subprocess.check_output.assert_called_with(
        ["docker", "--version"], env={"DOCKER_CLI_HINTS": "false"}
    )


def test_docker_failure(mock_tools, user_mapping_run_calls, capsys):
    """If docker failed during execution, the Docker wrapper is returned with a
    warning."""
    # Mock Docker cannot be found
    mock_tools.subprocess.check_output.side_effect = [
        subprocess.CalledProcessError(
            returncode=1,
            cmd="docker --version",
        ),
        "Success!",
        VALID_BUILDX_VERSION,
        VALID_USER_MAPPING_IMAGE_CACHE,
        "",  # touch host_write_file
        "",  # rm -f host_write_file
    ]

    # Invoke Docker verify
    result = Docker.verify(mock_tools)

    # The verify call should return the Docker wrapper
    assert isinstance(result, Docker)

    # Docker version and plugins were verified and user mapping inspection occurred
    mock_tools.subprocess.check_output.assert_has_calls(
        DOCKER_VERIFICATION_CALLS + user_mapping_run_calls
    )

    # console output
    output = capsys.readouterr()
    assert "** WARNING: Unable to determine if Docker is installed" in output.out
    assert output.err == ""


def test_docker_bad_version(mock_tools, capsys):
    """If docker exists but the version string doesn't make sense, the Docker wrapper is
    returned with a warning."""
    # Mock a bad return value of `docker --version`
    mock_tools.subprocess.check_output.return_value = "Docker version 19.03.8, build\n"

    # Invoke Docker verify
    with pytest.raises(
        BriefcaseCommandError,
        match=rf"Briefcase requires Docker {DOCKER_VERSION_MIN} or higher",
    ):
        Docker.verify(mock_tools)


def test_docker_unknown_version(mock_tools, user_mapping_run_calls, capsys):
    """If docker exists but the version string doesn't make sense, the Docker wrapper is
    returned with a warning."""
    # Mock a bad return value of `docker --version`
    mock_tools.subprocess.check_output.return_value = "ceci nest pas un Docker\n"

    # Invoke Docker verify
    result = Docker.verify(mock_tools)

    # The verify call should return the Docker wrapper
    assert isinstance(result, Docker)

    # Docker version and plugins were verified and user mapping inspection occurred
    mock_tools.subprocess.check_output.assert_has_calls(
        DOCKER_VERIFICATION_CALLS + user_mapping_run_calls
    )

    # console output
    output = capsys.readouterr()
    assert "** WARNING: Unable to determine the version of Docker" in output.out
    assert output.err == ""


def test_docker_exists_but_process_lacks_permission_to_use_it(mock_tools):
    """If the docker daemon isn't running, the check fails."""
    error_message = """
Client:
 Debug Mode: false

Server:
ERROR: Got permission denied while trying to connect to the Docker daemon socket at unix:///var/run/docker.sock:

Get http://%2Fvar%2Frun%2Fdocker.sock/v1.40/info: dial unix /var/run/docker.sock: connect: permission denied
errors pretty printing info"""

    mock_tools.subprocess.check_output.side_effect = [
        VALID_DOCKER_VERSION,
        subprocess.CalledProcessError(
            returncode=1,
            cmd="docker info",
            output=error_message,
        ),
    ]
    with pytest.raises(
        BriefcaseCommandError,
        match="does not have\npermissions to invoke Docker.",
    ):
        Docker.verify(mock_tools)


@pytest.mark.parametrize(
    "error_message",
    [
        """
Client:
 Debug Mode: false

Server:
ERROR: Error response from daemon: dial unix docker.raw.sock: connect: connection refused
errors pretty printing info
""",  # this is the error shown on mac
        """
Client:
 Debug Mode: false

Server:
ERROR: Cannot connect to the Docker daemon at unix:///var/run/docker.sock. Is the docker daemon running?
errors pretty printing info""",  # this is the error show on linux
    ],
)
def test_docker_exists_but_is_not_running(error_message, mock_tools):
    """If the docker daemon isn't running, the check fails."""
    mock_tools.subprocess.check_output.side_effect = [
        VALID_DOCKER_VERSION,
        subprocess.CalledProcessError(
            returncode=1,
            cmd="docker info",
            output=error_message,
        ),
    ]
    with pytest.raises(
        BriefcaseCommandError,
        match="the Docker\ndaemon is not running",
    ):
        Docker.verify(mock_tools)


def test_docker_exists_but_unknown_error_when_running_command(mock_tools):
    """If docker info fails in unknown ways, the check fails."""
    mock_tools.subprocess.check_output.side_effect = [
        VALID_DOCKER_VERSION,
        subprocess.CalledProcessError(
            returncode=1,
            cmd="docker info",
            output="This command failed!",
        ),
    ]

    with pytest.raises(
        BriefcaseCommandError,
        match="Check your Docker\ninstallation, and try again",
    ):
        Docker.verify(mock_tools)


def test_buildx_plugin_not_installed(mock_tools):
    """If the buildx plugin is not installed, verification fails."""
    mock_tools.subprocess.check_output.side_effect = [
        VALID_DOCKER_VERSION,
        VALID_DOCKER_INFO,
        subprocess.CalledProcessError(
            returncode=1,
            cmd="docker buildx version",
        ),
    ]

    with pytest.raises(
        BriefcaseCommandError,
        match="Docker is installed and available for use but the buildx plugin\nis not installed",
    ):
        Docker.verify(mock_tools)


def test_docker_image_hint(mock_tools):
    """If an image_tag is passed to verification, it is used for the user mapping
    check."""
    # Mock the return values for Docker verification
    mock_tools.subprocess.check_output.side_effect = [
        VALID_DOCKER_VERSION,
        VALID_DOCKER_INFO,
        VALID_BUILDX_VERSION,
        VALID_USER_MAPPING_IMAGE_CACHE,
        "",  # touch host_write_file
        "",  # rm -f host_write_file
    ]

    Docker.verify(mock_tools, image_tag="myimage:tagtorulethemall")

    mock_tools.subprocess.check_output.assert_has_calls(
        DOCKER_VERIFICATION_CALLS
        + [
            call(
                ["docker", "images", "-q", "myimage:tagtorulethemall"],
                env={"DOCKER_CLI_HINTS": "false"},
            ),
            call(
                args=[
                    "docker",
                    "run",
                    "--rm",
                    "--volume",
                    f"{Path.cwd() / 'build'}:/host_write_test:z",
                    "myimage:tagtorulethemall",
                    "touch",
                    "/host_write_test/container_write_test",
                ],
                env={"DOCKER_CLI_HINTS": "false"},
            ),
            call(
                args=[
                    "docker",
                    "run",
                    "--rm",
                    "--volume",
                    f"{Path.cwd() / 'build'}:/host_write_test:z",
                    "myimage:tagtorulethemall",
                    "rm",
                    "-f",
                    "/host_write_test/container_write_test",
                ],
                env={"DOCKER_CLI_HINTS": "false"},
            ),
        ]
    )


def test_user_mapping_write_file_path(mock_tools):
    """The write test file path is as expected."""
    expected_path = Path.cwd() / "build/container_write_test"
    assert Docker(mock_tools)._write_test_path() == expected_path


def test_user_mapping_write_file_exists(mock_tools, mock_write_test_path):
    """Docker verification fails when the container write test file exists and cannot be
    deleted."""
    # Mock the return values for Docker verification
    mock_tools.subprocess.check_output.side_effect = [
        VALID_DOCKER_VERSION,
        VALID_DOCKER_INFO,
        VALID_BUILDX_VERSION,
        VALID_USER_MAPPING_IMAGE_CACHE,
    ]

    # Mock failure for deleting an existing write test file
    mock_write_test_path.unlink = MagicMock(side_effect=OSError("delete failed"))

    # Fails when file cannot be deleted
    with pytest.raises(
        BriefcaseCommandError,
        match="file path used to determine how Docker is mapping users",
    ):
        Docker.verify(mock_tools)


def test_user_mapping_write_test_file_creation_fails(mock_tools, mock_write_test_path):
    """Docker verification fails if the write test file cannot be written."""
    # Mock the return values for Docker verification
    mock_tools.subprocess.check_output.side_effect = [
        VALID_DOCKER_VERSION,
        VALID_DOCKER_INFO,
        VALID_BUILDX_VERSION,
        VALID_USER_MAPPING_IMAGE_CACHE,
        # Mock failure for deleting an existing write test file
        subprocess.CalledProcessError(returncode=1, cmd=["docker", "run", "..."]),
    ]

    # Fails when file cannot be deleted
    with pytest.raises(
        BriefcaseCommandError,
        match="Unable to determine if Docker is mapping users",
    ):
        Docker.verify(mock_tools)


def test_user_mapping_write_test_file_cleanup_fails(mock_tools, mock_write_test_path):
    """Docker verification fails if the write test file cannot be removed after the
    test."""
    # Mock the return values for Docker verification
    mock_tools.subprocess.check_output.side_effect = [
        VALID_DOCKER_VERSION,
        VALID_DOCKER_INFO,
        VALID_BUILDX_VERSION,
        VALID_USER_MAPPING_IMAGE_CACHE,
        # Mock failure for deleting an existing write test file
        "",  # touch host_write_file
        subprocess.CalledProcessError(returncode=1, cmd=["docker", "run", "..."]),
    ]

    # Fails when file cannot be deleted
    with pytest.raises(
        BriefcaseCommandError,
        match="Unable to clean up from determining if Docker is mapping users",
    ):
        Docker.verify(mock_tools)


@pytest.mark.parametrize("file_owner_id, expected", [(1000, True), (0, False)])
def test_user_mapping_setting(
    mock_tools,
    user_mapping_run_calls,
    file_owner_id,
    expected,
):
    """If the write test file is not owned by root, user mapping is enabled, else
    disabled."""
    # Mock the return values for Docker verification
    mock_tools.subprocess.check_output.side_effect = [
        VALID_DOCKER_VERSION,
        VALID_DOCKER_INFO,
        VALID_BUILDX_VERSION,
        VALID_USER_MAPPING_IMAGE_CACHE,
        "",  # touch host_write_file
        "",  # rm -f host_write_file
    ]

    stat_result = namedtuple("stat_result", "st_uid")
    # Mock the os.stat call returning a file owned by file_owner_id
    mock_tools.os.stat = MagicMock(return_value=stat_result(st_uid=file_owner_id))

    docker = Docker.verify(mock_tools)

    # Docker user mapping inspection occurred
    mock_tools.subprocess.check_output.assert_has_calls(
        DOCKER_VERIFICATION_CALLS + user_mapping_run_calls
    )

    assert docker.is_user_mapped is expected



from __future__ import annotations

from pathlib import Path

from briefcase.exceptions import MissingToolError
from briefcase.integrations.base import ManagedTool, ToolCache


class RCEdit(ManagedTool):
    name = "rcedit"
    full_name = "RCEdit"
    supported_host_os = {"Windows"}

    @property
    def download_url(self) -> str:
        return (
            "https://github.com/electron/rcedit/releases/download/v2.0.0/rcedit-x64.exe"
        )

    @property
    def rcedit_path(self) -> Path:
        return self.tools.base_path / "rcedit-x64.exe"

    @classmethod
    def verify_install(cls, tools: ToolCache, install: bool = True, **kwargs) -> RCEdit:
        """Verify that rcedit is available.

        :param tools: ToolCache of available tools
        :param install: Should the tool be installed if it is not found?
        :returns: A valid rcedit tool wrapper. If rcedit is not available, and was not
            installed, raises MissingToolError.
        """
        # short circuit since already verified and available
        if hasattr(tools, "rcedit"):
            return tools.rcedit

        rcedit = RCEdit(tools=tools)

        if not rcedit.exists():
            if install:
                tools.logger.info(
                    "RCEdit was not found; downloading and installing...",
                    prefix=cls.name,
                )
                rcedit.install()
            else:
                raise MissingToolError("RCEdit")

        tools.rcedit = rcedit
        return rcedit

    def exists(self) -> bool:
        return self.rcedit_path.exists()

    def install(self):
        """Download and install RCEdit."""
        self.tools.file.download(
            url=self.download_url,
            download_path=self.tools.base_path,
            role="RCEdit",
        )

    def uninstall(self):
        """Uninstall RCEdit."""
        with self.tools.input.wait_bar("Removing old RCEdit install..."):
            self.rcedit_path.unlink()



import subprocess
from unittest.mock import MagicMock, Mock

import pytest

from briefcase.integrations.subprocess import Subprocess


@pytest.fixture
def mock_sub(mock_tools):
    mock_tools.input.release_console_control = Mock(
        wraps=mock_tools.input.release_console_control
    )
    mock_sub = Subprocess(mock_tools)
    mock_sub._subprocess = MagicMock(spec_set=subprocess)
    mock_sub._run_and_stream_output = MagicMock()
    return mock_sub


@pytest.mark.parametrize("batch_script", ["HELLO.BAT", "hello.bat", "hElLo.BaT"])
def test_run_windows_batch_script(mock_sub, batch_script):
    """Console control is released for a Windows batch script in run."""
    # Console control is only released on Windows
    mock_sub.tools.host_os = "Windows"

    with mock_sub.tools.input.wait_bar("Testing..."):
        mock_sub.run([batch_script, "World"])

    mock_sub._run_and_stream_output.assert_called_with(
        [batch_script, "World"],
        filter_func=None,
    )
    mock_sub.tools.input.release_console_control.assert_called_once()


@pytest.mark.parametrize("batch_script", ["HELLO.BAT", "hello.bat", "hElLo.BaT"])
def test_check_output_windows_batch_script(mock_sub, batch_script, sub_check_output_kw):
    """Console control is released for a Windows batch script in check_output."""
    # Console control is only released on Windows for batch scripts
    mock_sub.tools.host_os = "Windows"

    with mock_sub.tools.input.wait_bar("Testing..."):
        mock_sub.check_output([batch_script, "World"])

    mock_sub._subprocess.check_output.assert_called_with(
        [batch_script, "World"],
        **sub_check_output_kw,
    )
    mock_sub.tools.input.release_console_control.assert_called_once()


@pytest.mark.parametrize("sub_kwargs", [{"stream_output": True}, {}])
def test_run_stream_output_true(mock_sub, sub_kwargs):
    """Console control is not released when stream_output=True or is unspecified."""
    with mock_sub.tools.input.wait_bar("Testing..."):
        mock_sub.run(["Hello", "World"], **sub_kwargs)

    mock_sub._run_and_stream_output.assert_called_with(
        ["Hello", "World"],
        filter_func=None,
    )
    mock_sub.tools.input.release_console_control.assert_not_called()


def test_run_stream_output_false(mock_sub, sub_kw):
    """Console control is released when stream_output=False."""
    with mock_sub.tools.input.wait_bar("Testing..."):
        mock_sub.run(["Hello", "World"], stream_output=False)

    mock_sub._subprocess.run.assert_called_with(["Hello", "World"], **sub_kw)
    mock_sub.tools.input.release_console_control.assert_called_once()


@pytest.mark.parametrize(
    "cmdline, kwargs",
    [
        ([], dict()),
        (["Hello", "World"], dict()),
        (["Hello", "World"], dict(val1="value1", val2="value2")),
    ],
)
def test_negative_condition_not_controlled(
    mock_sub,
    cmdline,
    sub_check_output_kw,
    kwargs,
):
    """Passthrough to Subprocess if conditions to release console control are not met
    while the console is not controlled."""
    mock_sub.run(cmdline, **kwargs)
    mock_sub._run_and_stream_output.assert_called_with(
        cmdline,
        filter_func=None,
        **kwargs,
    )

    mock_sub.check_output(cmdline, **kwargs)

    final_kwargs = {**sub_check_output_kw, **kwargs}
    mock_sub._subprocess.check_output.assert_called_with(cmdline, **final_kwargs)
    mock_sub.tools.input.release_console_control.assert_not_called()


@pytest.mark.parametrize(
    "cmdline, kwargs",
    [
        ([], dict()),
        (["Hello", "World"], dict()),
        (["Hello", "World"], dict(val1="value1", val2="value2")),
    ],
)
def test_negative_condition_controlled(mock_sub, cmdline, kwargs, sub_check_output_kw):
    """Passthrough to Subprocess if conditions to release console control are not met
    while the console is controlled."""
    with mock_sub.tools.input.wait_bar("Testing..."):
        mock_sub.run(cmdline, **kwargs)
        mock_sub.check_output(cmdline, **kwargs)

    mock_sub._run_and_stream_output.assert_called_with(
        cmdline,
        filter_func=None,
        **kwargs,
    )
    final_kwargs = {**sub_check_output_kw, **kwargs}
    mock_sub._subprocess.check_output.assert_called_with(cmdline, **final_kwargs)
    mock_sub.tools.input.release_console_control.assert_not_called()



from subprocess import CalledProcessError
from unittest import mock

from briefcase.integrations.xcode import Xcode


def test_verify_xcode_install(mock_tools):
    """Xcode can be verified."""
    mock_tools.subprocess.check_output.side_effect = [
        "/Applications/Xcode/app/Contents/Developer",  # xcode-select -p
        "Xcode 14.0.1",  # xcodebuild -version
        CalledProcessError(cmd=["xcode-select", "--install"], returncode=1),
        "clang 37.42",  # clang --version
    ]

    Xcode.verify(mock_tools)

    # Both Xcode and the command line tools are verified
    assert mock_tools.xcode is not None
    assert mock_tools.xcode_cli is not None


def test_reverify_xcode_install(mock_tools):
    """A second call to verify is a no-op."""

    xcode = mock.MagicMock()
    mock_tools.xcode = xcode

    xcode_cli = mock.MagicMock()
    mock_tools.xcode_cli = xcode_cli

    Xcode.verify(mock_tools)

    # Both Xcode and the command line tools are verified
    assert mock_tools.xcode == xcode
    assert mock_tools.xcode_cli is not None



import sys
from unittest.mock import MagicMock, call

import pytest

from briefcase.console import Console, Log
from briefcase.integrations.subprocess import Subprocess
from briefcase.platforms.iOS.xcode import iOSXcodeUpdateCommand


@pytest.fixture
def update_command(tmp_path):
    return iOSXcodeUpdateCommand(
        logger=Log(),
        console=Console(),
        base_path=tmp_path / "base_path",
        data_path=tmp_path / "briefcase",
    )


def test_extra_pip_args(update_command, first_app_generated, tmp_path):
    """Extra iOS-specific args are included in calls to pip during update."""
    # Hard code the current architecture for testing. We only install simulator
    # requirements for the current platform.
    update_command.tools.host_arch = "wonky"

    first_app_generated.requires = ["something==1.2.3", "other>=2.3.4"]

    update_command.tools[first_app_generated].app_context = MagicMock(
        spec_set=Subprocess
    )

    update_command.install_app_requirements(first_app_generated, test_mode=False)

    bundle_path = tmp_path / "base_path/build/first-app/ios/xcode"
    assert update_command.tools[first_app_generated].app_context.run.mock_calls == [
        call(
            [
                sys.executable,
                "-u",
                "-X",
                "utf8",
                "-m",
                "pip",
                "install",
                "--disable-pip-version-check",
                "--no-python-version-warning",
                "--upgrade",
                "--no-user",
                f"--target={bundle_path / 'app_packages.iphoneos'}",
                "--prefer-binary",
                "--extra-index-url",
                "https://pypi.anaconda.org/beeware/simple",
                "something==1.2.3",
                "other>=2.3.4",
            ],
            check=True,
            encoding="UTF-8",
            env={
                "PYTHONPATH": str(
                    tmp_path
                    / "base_path"
                    / "build"
                    / "first-app"
                    / "ios"
                    / "xcode"
                    / "support"
                    / "platform-site"
                    / "iphoneos.arm64"
                )
            },
        ),
        call(
            [
                sys.executable,
                "-u",
                "-X",
                "utf8",
                "-m",
                "pip",
                "install",
                "--disable-pip-version-check",
                "--no-python-version-warning",
                "--upgrade",
                "--no-user",
                f"--target={bundle_path / 'app_packages.iphonesimulator'}",
                "--prefer-binary",
                "--extra-index-url",
                "https://pypi.anaconda.org/beeware/simple",
                "something==1.2.3",
                "other>=2.3.4",
            ],
            check=True,
            encoding="UTF-8",
            env={
                "PYTHONPATH": str(
                    tmp_path
                    / "base_path"
                    / "build"
                    / "first-app"
                    / "ios"
                    / "xcode"
                    / "support"
                    / "platform-site"
                    / "iphonesimulator.wonky"
                )
            },
        ),
    ]



from unittest.mock import MagicMock

import pytest

from briefcase.integrations.base import ToolCache
from briefcase.integrations.file import File
from briefcase.integrations.flatpak import Flatpak
from briefcase.integrations.subprocess import Subprocess


@pytest.fixture
def mock_tools(mock_tools) -> ToolCache:
    mock_tools.host_os = "Linux"
    mock_tools.host_arch = "gothic"

    # Mock default tools
    mock_tools.subprocess = MagicMock(spec_set=Subprocess)
    mock_tools.file.download = MagicMock(spec_set=File.download)

    return mock_tools


@pytest.fixture
def flatpak(mock_tools):
    return Flatpak(tools=mock_tools)



from unittest import mock

import pytest

from briefcase.commands.base import BaseCommand
from briefcase.console import Console, Log
from briefcase.exceptions import BriefcaseCommandError, InvalidDeviceError
from briefcase.platforms.iOS.xcode import iOSXcodeMixin
from tests.utils import DummyConsole


class DummyCommand(iOSXcodeMixin, BaseCommand):
    """A dummy command that includes the iOS Xcode mixin."""

    command = "dummy"

    def __init__(self, base_path, **kwargs):
        kwargs.setdefault("logger", Log())
        kwargs.setdefault("console", Console())
        super().__init__(base_path=base_path, **kwargs)
        self.tools.input = DummyConsole()


@pytest.fixture
def dummy_command(tmp_path):
    cmd = DummyCommand(base_path=tmp_path)

    # Mock the options object
    cmd.options = mock.MagicMock()
    cmd.options.device = None

    # Mock get_simulators
    mock_get_simulators = mock.MagicMock()
    cmd.get_simulators = mock_get_simulators

    return cmd


def test_explicit_device_udid(dummy_command):
    """If the user nominates a device UDID at the command line, it is used."""
    # get_simulators will return some options.
    dummy_command.get_simulators.return_value = {
        "iOS 13.2": {
            "C9A005C8-9468-47C5-8376-68A6E3408209": "iPhone 8",
            "2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D": "iPhone 11",
            "EEEBA06C-81F9-407C-885A-2261306DB2BE": "iPhone 11 Pro Max",
        }
    }

    # The target device will be the one the user specified as an option.
    result = dummy_command.select_target_device("2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D")
    udid, iOS_version, device = result

    assert udid == "2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D"
    assert iOS_version == "13.2"
    assert device == "iPhone 11"


def test_explicit_device_name_should_find_highest_version(dummy_command):
    """If an explicit device name is provided (case insensitive) it is used."""
    # get_simulators will return multiple options on 2 iOS versions.
    dummy_command.get_simulators.return_value = {
        "iOS 10.3": {
            "0BB80120-FA02-4597-A1BA-DB8CDE4F086D": "iPhone 5s",
            "D7BBAD14-38FD-48F5-ACFD-B1193F829216": "iPhone 6",
            "6998CA09-44B5-4963-8F80-265412D99683": "iPhone 7",
            "68A6E340-8376-47C5-9468-C9A005C88209": "iPhone 8",
        },
        "iOS 13.2": {
            "C9A005C8-9468-47C5-8376-68A6E3408209": "iPhone 8",
            "2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D": "iPhone 11",
            "EEEBA06C-81F9-407C-885A-2261306DB2BE": "iPhone 11 Pro Max",
        },
    }

    # Try to load an iPhone 8.
    # It matches case insensitive, and finds the highest iOS version.
    udid, iOS_version, device = dummy_command.select_target_device("iphone 8")

    assert udid == "C9A005C8-9468-47C5-8376-68A6E3408209"
    assert iOS_version == "13.2"
    assert device == "iPhone 8"

    # User input was not solicited
    assert dummy_command.input.prompts == []


def test_explicit_device_name_should_find_highest_version_no_os(dummy_command):
    """If an explicit device name is provided (case insensitive) it is used."""
    # get_simulators will return multiple options on 2 iOS versions.
    # In older versions of Xcode, the OS name wasn't included in the version
    # string. Ensure that this still works, even though it doesn't happen in
    # practice (as of May 2022)
    dummy_command.get_simulators.return_value = {
        "10.3": {
            "0BB80120-FA02-4597-A1BA-DB8CDE4F086D": "iPhone 5s",
            "D7BBAD14-38FD-48F5-ACFD-B1193F829216": "iPhone 6",
            "6998CA09-44B5-4963-8F80-265412D99683": "iPhone 7",
            "68A6E340-8376-47C5-9468-C9A005C88209": "iPhone 8",
        },
        "13.2": {
            "C9A005C8-9468-47C5-8376-68A6E3408209": "iPhone 8",
            "2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D": "iPhone 11",
            "EEEBA06C-81F9-407C-885A-2261306DB2BE": "iPhone 11 Pro Max",
        },
    }

    # Try to load an iPhone 8.
    # It matches case insensitive, and finds the highest iOS version.
    udid, iOS_version, device = dummy_command.select_target_device("iphone 8")

    assert udid == "C9A005C8-9468-47C5-8376-68A6E3408209"
    assert iOS_version == "13.2"
    assert device == "iPhone 8"

    # User input was not solicited
    assert dummy_command.input.prompts == []


def test_explicit_device_name_and_version(dummy_command):
    """If there are multiple options on multiple devices, two user inputs are needed."""
    # get_simulators will return multiple options on 2 iOS versions.
    dummy_command.get_simulators.return_value = {
        "iOS 10.3": {
            "0BB80120-FA02-4597-A1BA-DB8CDE4F086D": "iPhone 5s",
            "D7BBAD14-38FD-48F5-ACFD-B1193F829216": "iPhone 6",
            "6998CA09-44B5-4963-8F80-265412D99683": "iPhone 7",
            "68A6E340-8376-47C5-9468-C9A005C88209": "iPhone 8",
        },
        "iOS 13.2": {
            "C9A005C8-9468-47C5-8376-68A6E3408209": "iPhone 8",
            "2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D": "iPhone 11",
            "EEEBA06C-81F9-407C-885A-2261306DB2BE": "iPhone 11 Pro Max",
        },
    }

    # Specify an iPhone 8 on 10.3
    # It matches case insensitive, and finds the explicit iOS version.
    udid, iOS_version, device = dummy_command.select_target_device("iphone 8::ios 10.3")

    assert udid == "68A6E340-8376-47C5-9468-C9A005C88209"
    assert iOS_version == "10.3"
    assert device == "iPhone 8"

    # User input was not solicited
    assert dummy_command.input.prompts == []


def test_invalid_explicit_device_udid(dummy_command):
    """If the user nominates an invalid device UDID, an error is raised."""
    # get_simulators will some options.
    dummy_command.get_simulators.return_value = {
        "iOS 13.2": {
            "C9A005C8-9468-47C5-8376-68A6E3408209": "iPhone 8",
            "2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D": "iPhone 11",
            "EEEBA06C-81F9-407C-885A-2261306DB2BE": "iPhone 11 Pro Max",
        }
    }

    # The user nominates a specific device that doesn't exist
    with pytest.raises(InvalidDeviceError):
        dummy_command.select_target_device("deadbeef-dead-beef-cafe-deadbeefdead")

    # User input was not solicited
    assert dummy_command.input.prompts == []


def test_invalid_explicit_device_name(dummy_command):
    """If the user nominates an invalid device name, an error is raised."""
    # get_simulators will some options.
    dummy_command.get_simulators.return_value = {
        "iOS 13.2": {
            "C9A005C8-9468-47C5-8376-68A6E3408209": "iPhone 8",
            "2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D": "iPhone 11",
            "EEEBA06C-81F9-407C-885A-2261306DB2BE": "iPhone 11 Pro Max",
        }
    }

    # The user nominates a specific device name that doesn't exist
    with pytest.raises(InvalidDeviceError):
        dummy_command.select_target_device("iphone 99")

    # User input was not solicited
    assert dummy_command.input.prompts == []


def test_explicit_name_invalid_version(dummy_command):
    """If the user nominates a device name but an invalid version, an error is
    raised."""
    # get_simulators will some options.
    dummy_command.get_simulators.return_value = {
        "iOS 13.2": {
            "C9A005C8-9468-47C5-8376-68A6E3408209": "iPhone 8",
            "2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D": "iPhone 11",
            "EEEBA06C-81F9-407C-885A-2261306DB2BE": "iPhone 11 Pro Max",
        }
    }

    # The user nominates a valid device name, but on a version that
    # doesn't exist.
    with pytest.raises(InvalidDeviceError):
        dummy_command.select_target_device("iphone 11::37.42")


def test_invalid_explicit_device_name_valid_version(dummy_command):
    """If the user nominates an invalid device name but a valid version, an error is
    raised."""
    # get_simulators will some options.
    dummy_command.get_simulators.return_value = {
        "iOS 13.2": {
            "C9A005C8-9468-47C5-8376-68A6E3408209": "iPhone 8",
            "2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D": "iPhone 11",
            "EEEBA06C-81F9-407C-885A-2261306DB2BE": "iPhone 11 Pro Max",
        }
    }

    # The user nominates a valid device name, but on a version that
    # doesn't exist.
    with pytest.raises(InvalidDeviceError):
        dummy_command.select_target_device("iphone 99::iOS 13.2")


def test_implied_device(dummy_command):
    """If there's only one device, no input is required; the device is returned."""
    # get_simulators will return one option.
    dummy_command.get_simulators.return_value = {
        "iOS 13.2": {
            "2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D": "iPhone 11",
        }
    }

    udid, iOS_version, device = dummy_command.select_target_device()

    assert udid == "2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D"
    assert iOS_version == "13.2"
    assert device == "iPhone 11"

    # No user input was solicited
    assert dummy_command.input.prompts == []


def test_implied_os(dummy_command):
    """If there is only one OS option, it's implied."""
    # get_simulators will return multiple options on 1 iOS version.
    dummy_command.get_simulators.return_value = {
        "iOS 13.2": {
            "C9A005C8-9468-47C5-8376-68A6E3408209": "iPhone 8",
            "2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D": "iPhone 11",
            "EEEBA06C-81F9-407C-885A-2261306DB2BE": "iPhone 11 Pro Max",
        }
    }

    # Return option 2 (iPhone 11)
    dummy_command.input.values = ["2"]

    udid, iOS_version, device = dummy_command.select_target_device()

    assert udid == "EEEBA06C-81F9-407C-885A-2261306DB2BE"
    assert iOS_version == "13.2"
    assert device == "iPhone 11 Pro Max"

    # User input was solicited once
    assert dummy_command.input.prompts == ["> "]


def test_multiple_os_implied_device(dummy_command):
    """If there are multiple OS options, but only one device on the chosen OS, device is
    implied."""
    # get_simulators will return multiple options on 1 iOS version.
    dummy_command.get_simulators.return_value = {
        "iOS 10.3": {
            "0BB80120-FA02-4597-A1BA-DB8CDE4F086D": "iPhone 5s",
            "D7BBAD14-38FD-48F5-ACFD-B1193F829216": "iPhone 6",
            "6998CA09-44B5-4963-8F80-265412D99683": "iPhone 7",
        },
        "iOS 13.2": {
            "2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D": "iPhone 11",
        },
    }

    # Return option 1 (13.2)
    dummy_command.input.values = ["2"]

    # Device for iOS 13.2 is implied.
    udid, iOS_version, device = dummy_command.select_target_device()

    assert udid == "2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D"
    assert iOS_version == "13.2"
    assert device == "iPhone 11"

    # User input was solicited once
    assert dummy_command.input.prompts == ["> "]


def test_os_and_device_options(dummy_command):
    """If there are multiple options on multiple devices, two user inputs are needed."""
    # get_simulators will return multiple options on 2 iOS versions.
    dummy_command.get_simulators.return_value = {
        "iOS 10.3": {
            "0BB80120-FA02-4597-A1BA-DB8CDE4F086D": "iPhone 5s",
            "D7BBAD14-38FD-48F5-ACFD-B1193F829216": "iPhone 6",
            "6998CA09-44B5-4963-8F80-265412D99683": "iPhone 7",
        },
        "iOS 13.2": {
            "C9A005C8-9468-47C5-8376-68A6E3408209": "iPhone 8",
            "2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D": "iPhone 11",
            "EEEBA06C-81F9-407C-885A-2261306DB2BE": "iPhone 11 Pro Max",
        },
    }

    # Return option 2 (13.2), then option 1 (iPhone 11)
    dummy_command.input.values = ["2", "1"]

    udid, iOS_version, device = dummy_command.select_target_device()

    assert udid == "2D3503A3-6EB9-4B37-9B17-C7EFEF2FA32D"
    assert iOS_version == "13.2"
    assert device == "iPhone 11"

    # User input was solicited twice
    assert dummy_command.input.prompts == ["> "] * 2


def test_no_os_versions(dummy_command):
    """If there are no supported OS versions, raise an error."""
    # get_simulators returns no valid iOS versions.
    dummy_command.get_simulators.return_value = {}

    with pytest.raises(BriefcaseCommandError):
        dummy_command.select_target_device()


def test_no_devices_for_os(dummy_command):
    """If there are no devices for an OS version, raise an error."""
    # get_simulators returns no devices for iOS 13.2
    dummy_command.get_simulators.return_value = {"iOS 13.2": {}}

    with pytest.raises(BriefcaseCommandError):
        dummy_command.select_target_device()



from unittest.mock import MagicMock

import pytest

from briefcase.integrations.cookiecutter import XMLExtension


@pytest.mark.parametrize(
    "value, expected",
    [
        # Literal booleans
        (True, "true"),
        (False, "false"),
        # True-ish values
        (1, "true"),
        ("Hello", "true"),
        # False-ish values
        (0, "false"),
        ("", "false"),
    ],
)
def test_bool_attr(value, expected):
    env = MagicMock()
    env.filters = {}
    XMLExtension(env)
    assert env.filters["bool_attr"](value) == expected



from __future__ import annotations

import re
import sys
import unicodedata
from collections import OrderedDict
from collections.abc import Sequence
from email.utils import parseaddr
from typing import Iterable
from urllib.parse import urlparse

if sys.version_info >= (3, 10):  # pragma: no-cover-if-lt-py310
    from importlib.metadata import entry_points
else:  # pragma: no-cover-if-gte-py310
    # Before Python 3.10, entry_points did not support the group argument;
    # so, the backport package must be used on older versions.
    from importlib_metadata import entry_points

from briefcase.bootstraps import BaseGuiBootstrap
from briefcase.config import (
    is_valid_app_name,
    is_valid_bundle_identifier,
    make_class_name,
)
from briefcase.console import MAX_TEXT_WIDTH
from briefcase.console import select_option as _select_option
from briefcase.exceptions import BriefcaseCommandError
from briefcase.integrations.git import Git

from .base import BaseCommand


def titlecase(s):
    """Convert a string to titlecase.

    Follow Chicago Manual of Style rules for capitalization (roughly).

    * Capitalize *only* the first letter of each word
    * ... unless the word is an acronym (e.g., URL)
    * ... or the word is on the exclude list ('of', 'and', 'the')
    :param s: The input string
    :returns: A capitalized string.
    """
    return " ".join(
        (
            word
            if (
                word.isupper()
                or word
                in {
                    "a",
                    "an",
                    "and",
                    "as",
                    "at",
                    "but",
                    "by",
                    "en",
                    "for",
                    "if",
                    "in",
                    "of",
                    "on",
                    "or",
                    "the",
                    "to",
                    "via",
                    "vs",
                }
            )
            else word.capitalize()
        )
        for word in s.split(" ")
    )


def get_gui_bootstraps() -> dict[str, type[BaseGuiBootstrap]]:
    """Loads built-in and third-party GUI bootstraps."""
    return {
        entry_point.name: entry_point.load()
        for entry_point in entry_points(group="briefcase.bootstraps")
    }


def parse_project_overrides(project_overrides: list[str]) -> dict[str, str]:
    """Parse the command-line arguments to override new project defaults."""
    overrides = {}

    if project_overrides:
        for override in project_overrides:
            try:
                key, value = override.split("=", 1)
            except ValueError as e:
                raise BriefcaseCommandError(
                    f"Unable to parse project configuration override {override!r}"
                ) from e
            else:
                if not ((key := key.strip()) and (value := value.strip())):
                    raise BriefcaseCommandError(
                        f"Invalid Project configuration override {override!r}"
                    )
                if key in overrides:
                    raise BriefcaseCommandError(
                        f"Project configuration override {key!r} specified multiple times"
                    )
                overrides[key] = value
    return overrides


class NewCommand(BaseCommand):
    cmd_line = "briefcase new"
    command = "new"
    platform = "all"
    output_format = None
    description = "Create a new Briefcase project."

    def bundle_path(self, app):
        """A placeholder; New command doesn't have a bundle path."""
        raise NotImplementedError()

    def binary_path(self, app):
        """A placeholder; New command doesn't have a binary path."""
        raise NotImplementedError()

    def parse_config(self, filename, overrides):
        """There is no configuration when starting a new project; this implementation
        overrides the base so that no config is parsed."""
        pass

    def add_options(self, parser):
        parser.add_argument(
            "-t",
            "--template",
            dest="template",
            help="The cookiecutter template to use for the new project",
        )

        parser.add_argument(
            "--template-branch",
            dest="template_branch",
            help="The branch of the cookiecutter template to use for the new project",
        )

        parser.add_argument(
            "-Q",
            dest="project_overrides",
            action="append",
            metavar="KEY=VALUE",
            help="Override the value of the project configuration item KEY with VALUE",
        )

    def make_app_name(self, formal_name):
        """Construct a candidate app name from a formal name.

        :param formal_name: The formal name
        :returns: The candidate app name
        """
        normalized = unicodedata.normalize("NFKD", formal_name)
        stripped = re.sub("[^0-9a-zA-Z_]+", "", normalized).lstrip("_")
        if stripped:
            return stripped.lower()
        else:
            # If stripping removes all the content,
            # use a dummy app name as the suggestion.
            return "myapp"

    def validate_formal_name(self, candidate):
        """Determine if the formal name is valid.

        A formal name is valid if it contains at least one identifier character.

        :param candidate: The candidate name
        :returns: True the formal name is valid.
        :raises: ValueError if the name is not a valid formal name.
        """
        if not make_class_name(candidate):  # Check whether a class name may be derived
            raise ValueError(
                self.input.textwrap(
                    f"{candidate!r} is not a valid formal name.\n"
                    "\n"
                    "Formal names must include at least one valid Python identifier character."
                )
            )

        return True

    def _validate_existing_app_name(self, candidate):
        """Perform internal validation preventing the use of pre-existing app names.

        Invoked by validate_app_name; subclasses may override this behavior.
        """
        if (self.base_path / candidate).exists():
            raise ValueError(
                self.input.textwrap(
                    f"A {candidate!r} directory already exists.\n"
                    f"\n"
                    f"Select a different name, move to a different parent directory, or "
                    f"delete the existing folder."
                )
            )

    def validate_app_name(self, candidate):
        """Determine if the app name is valid.

        :param candidate: The candidate name
        :returns: True. If there are any validation problems, raises ValueError with a
            diagnostic message.
        """
        if not is_valid_app_name(candidate):
            raise ValueError(
                self.input.textwrap(
                    f"{candidate!r} is not a valid app name.\n"
                    "\n"
                    "App names must not be reserved keywords such as 'and', 'for' and "
                    "'while'. They must also be PEP508 compliant (i.e., they can only "
                    "include letters, numbers, '-' and '_'; must start with a letter; "
                    "and cannot end with '-' or '_')."
                )
            )

        # Validate if the existing app name
        self._validate_existing_app_name(candidate)

        return True

    def make_module_name(self, app_name):
        """Construct a valid module name from an app name.

        :param app_name: The app name
        :returns: The app's module name.
        """
        return app_name.replace("-", "_")

    def validate_bundle(self, candidate):
        """Determine if the bundle identifier is valid.

        :param candidate: The candidate bundle identifier
        :returns: True. If there are any validation problems, raises ValueError with a
            diagnostic message.
        """
        if not is_valid_bundle_identifier(candidate):
            raise ValueError(
                self.input.textwrap(
                    f"{candidate!r} is not a valid bundle identifier.\n"
                    "\n"
                    "The bundle should be a reversed domain name. It must contain at "
                    "least 2 dot-separated sections; each section may only include "
                    "letters, numbers, and hyphens; and each section may not contain any "
                    "reserved words (like 'switch', or 'while')."
                )
            )

        return True

    def make_domain(self, bundle):
        """Construct a candidate domain from a bundle identifier.

        :param bundle: The bundle identifier
        :returns: The candidate domain
        """
        return ".".join(bundle.split(".")[::-1])

    def make_author_email(self, author, bundle):
        """Construct a candidate email address from the authors name and the bundle
        identifier.

        The candidate is based on the assumption that the author's name is in
        "first/last" format, or it a corporate name; the "first" part is split off, and
        prepended to the domain extracted from the bundle.

        It's not a perfect system, but it's better than putting up "me@example.com" as a
        candidate default value.

        :param author: The authors name.
        :param bundle: The bundle identifier.
        :returns: The candidate author's name
        """
        return f"{author.split(' ')[0].lower()}@{self.make_domain(bundle)}"

    def validate_email(self, candidate):
        """Determine if the email address is valid.

        :param candidate: The candidate email address
        :returns: True. If there are any validation problems, raises ValueError with a
            diagnostic message.
        """
        if parseaddr(candidate)[1] != candidate:
            raise ValueError("Not a valid email address")
        return True

    def make_project_url(self, bundle, app_name):
        """Construct a candidate project URL from the bundle and app name.

        It's not a perfect guess, but it's better than having
        "https://example.com".

        :param bundle: The bundle identifier.
        :param app_name: The app name.
        :returns: The candidate project URL
        """
        return f"https://{self.make_domain(bundle)}/{app_name}"

    def validate_url(self, candidate):
        """Determine if the URL is valid.

        :param candidate: The candidate URL
        :returns: True. If there are any validation problems, raises ValueError with a
            diagnostic message.
        """
        result = urlparse(candidate)
        if not all([result.scheme, result.netloc]):
            raise ValueError("Not a valid URL!")
        if result.scheme not in {"http", "https"}:
            raise ValueError("Not a valid website URL!")
        return True

    def prompt_divider(self, title: str = ""):
        """Writes a divider with an optional title."""
        title = f"-- {title} " if title else ""
        self.input.prompt()
        self.input.prompt()
        self.input.prompt(f"{title}{'-' * (MAX_TEXT_WIDTH - len(title))}", style="bold")

    def prompt_intro(self, intro: str):
        """Write the introduction for a prompt."""
        self.input.prompt()
        self.input.prompt(self.input.textwrap(intro))
        self.input.prompt()

    def validate_user_input(self, validator, answer) -> bool:
        """Validate a user's input is acceptable.

        A warning message is issued if input is enabled. Otherwise, an exception is
        raised and project creation is aborted.
        """
        error_msg = f"Invalid value; {answer}"
        try:
            if validator is None or validator(answer):
                return True
        except ValueError as e:
            error_msg = str(e)

        if not self.input.enabled:
            raise BriefcaseCommandError(error_msg)
        self.logger.warning()
        self.logger.warning(error_msg)

        return False

    def validate_override(self, override_value: str | None, validator=None) -> bool:
        """Validates and returns override value for a project configuration."""
        if override_value is not None:
            self.input.prompt()
            self.input.prompt(f"Using override value {override_value!r}")
            if not self.validate_user_input(validator, override_value):
                return False
        return True

    def validate_selection_override(
        self,
        choices: Iterable[str],
        override_value: str | None,
    ) -> bool:
        """Validate a project override value against a list of options."""
        return self.validate_override(override_value, validator=lambda c: c in choices)

    def input_text(self, intro, variable, default, validator=None, override_value=None):
        """Read a text answer from the user.

        :param intro: An introductory paragraph explaining the question being asked.
        :param variable: The name of the variable being entered.
        :param default: The default value if the user hits enter without typing anything.
        :param validator: (optional) A validator function; accepts a single input (the
            candidate response), returns True if the answer is valid, or raises
            ValueError() with a debugging message if the candidate value isn't valid.
        :param override_value: value to return instead of soliciting input
        :returns: a string, guaranteed to meet the validation criteria of ``validator``
        """
        variable = titlecase(variable)
        self.prompt_divider(title=variable)

        if override_value and self.validate_override(override_value, validator):
            return override_value

        self.prompt_intro(intro=intro)
        while True:
            answer = self.input.text_input(f"{variable} [{default}]: ", default=default)

            if self.validate_user_input(validator, answer):
                return answer

            self.input.prompt()

    def select_option(
        self,
        intro: str,
        variable: str,
        default: str | None,
        options: Sequence[str],
        override_value: str | None,
    ) -> str:
        variable = titlecase(variable)
        self.prompt_divider(title=variable)

        if override_value is not None:
            if self.validate_selection_override(options, override_value):
                return override_value
            else:
                self.logger.warning(
                    f"Invalid override value {override_value!r} for {variable}, using user-provided value."
                )

        if default is not None:
            default = str(options.index(default) + 1)
        else:
            default = "1"

        self.prompt_intro(intro=intro)
        return _select_option(
            prompt=f"{variable} [{default}]:",
            input=self.input,
            default=default,
            options=list(zip(options, options)),
        )

    def input_project_name(self, formal_name, override_value):
        return self.input_text(
            intro=(
                "Briefcase can manage projects that contain multiple applications, so "
                "we need a Project name.\n"
                "\n"
                "If you're only planning to have one application in this project, you "
                "can use the formal name as the project name."
            ),
            variable="project name",
            default=formal_name,
            override_value=override_value,
        )

    def input_license(self, override_value: str | None):
        licenses = [
            "BSD license",
            "MIT license",
            "Apache Software License",
            "GNU General Public License v2 (GPLv2)",
            "GNU General Public License v2 or later (GPLv2+)",
            "GNU General Public License v3 (GPLv3)",
            "GNU General Public License v3 or later (GPLv3+)",
            "Proprietary",
            "Other",
        ]
        return self.select_option(
            intro="What license do you want to use for this project's code?",
            variable="Project License",
            options=licenses,
            default=None,
            override_value=override_value,
        )

    def build_context(self, project_overrides: dict[str, str]) -> dict[str, str]:
        """Builds the cookiecutter context dict for the new project."""
        context = self.build_app_context(project_overrides)
        context.update(self.build_gui_context(context, project_overrides))

        return context

    def build_app_context(self, project_overrides: dict[str, str]) -> dict[str, str]:
        """Ask the user for details about the app to be created.

        :returns: A context dictionary to be used in the cookiecutter project template.
        """
        formal_name = self.input_text(
            intro=(
                "First, we need a formal name for your application.\n"
                "\n"
                "This is the name that will be displayed to humans whenever the name "
                "of the application is displayed. It can have spaces and punctuation "
                "if you like, and any capitalization will be used as you type it."
            ),
            variable="formal name",
            default="Hello World",
            validator=self.validate_formal_name,
            override_value=project_overrides.pop("formal_name", None),
        )

        # The class name can be completely derived from the formal name.
        class_name = make_class_name(formal_name)

        default_app_name = self.make_app_name(formal_name)
        app_name = self.input_text(
            intro=(
                "Next, we need a name that can serve as a machine-readable Python "
                "package name for your application.\n"
                "\n"
                "This name must be PEP508-compliant - that means the name may only "
                "contain letters, numbers, hyphens and underscores; it can't contain "
                "spaces or punctuation, and it can't start with a hyphen or "
                "underscore.\n"
                "\n"
                "Based on your formal name, we suggest an app name of "
                f"{default_app_name!r}, but you can use another name if you want."
            ),
            variable="app name",
            default=default_app_name,
            validator=self.validate_app_name,
            override_value=project_overrides.pop("app_name", None),
        )

        # The module name can be completely derived from the app name.
        module_name = self.make_module_name(app_name)
        source_dir = f"src/{module_name}"
        test_source_dir = "tests"

        bundle = self.input_text(
            intro=(
                "Now we need a bundle identifier for your application.\n"
                "\n"
                "App stores need to protect against having multiple applications with "
                "the same name; the bundle identifier is the namespace they use to "
                "identify applications that come from you. The bundle identifier is "
                "usually the domain name of your company or project, in reverse order.\n"
                "\n"
                "For example, if you are writing an application for Example Corp, "
                "whose website is example.com, your bundle would be 'com.example'. "
                "The bundle will be combined with your application's machine readable "
                "name to form a complete application identifier (e.g., "
                f"com.example.{app_name})."
            ),
            variable="bundle identifier",
            default="com.example",
            validator=self.validate_bundle,
            override_value=project_overrides.pop("bundle", None),
        )

        project_name = self.input_project_name(
            formal_name, project_overrides.pop("project_name", None)
        )

        description = self.input_text(
            intro="Now, we need a one line description for your application.",
            variable="description",
            default="My first application",
            override_value=project_overrides.pop("description", None),
        )

        author = self.input_text(
            intro=(
                "Who do you want to be credited as the author of this application?\n"
                "\n"
                "This could be your own name, or the name of your company you work for."
            ),
            variable="author",
            default="Jane Developer",
            override_value=project_overrides.pop("author", None),
        )

        author_email = self.input_text(
            intro=(
                "What email address should people use to contact the developers of "
                "this application?\n"
                "\n"
                "This might be your own email address, or a generic contact address "
                "you set up specifically for this application."
            ),
            variable="author's email",
            default=self.make_author_email(author, bundle),
            validator=self.validate_email,
            override_value=project_overrides.pop("author_email", None),
        )

        url = self.input_text(
            intro=(
                "What is the website URL for this application?\n"
                "\n"
                "If you don't have a website set up yet, you can put in a dummy URL."
            ),
            variable="application URL",
            default=self.make_project_url(bundle, app_name),
            validator=self.validate_url,
            override_value=project_overrides.pop("url", None),
        )
        project_license = self.input_license(
            override_value=project_overrides.pop("license", None)
        )

        return {
            "formal_name": formal_name,
            "app_name": app_name,
            "class_name": class_name,
            "module_name": module_name,
            "source_dir": source_dir,
            "test_source_dir": test_source_dir,
            "project_name": project_name,
            "description": description,
            "author": author,
            "author_email": author_email,
            "bundle": bundle,
            "url": url,
            "license": project_license,
        }

    def build_gui_context(
        self,
        context: dict[str, str],
        project_overrides: dict[str, str],
    ) -> dict[str, str]:
        """Build context specific to the GUI toolkit."""
        bootstraps = get_gui_bootstraps()
        bootstrap_options = self._gui_bootstrap_choices(bootstraps)

        # Map the override value to the annotated override value so we can use it with self.select_option
        # If a user specifies the override value PySide6, then we want it to be mapped to the annotated value
        # "PySide6       (does not support iOS/Android deployment)"
        # since that is the option presented to the user. To accomplish this, we map the bootstrap classes
        # to their annotated names and do a reverse lookup from the override bootstrap class to the annotated name.
        #
        # We do it this way to ensure consistent prompting for the user.
        if bootstrap_override := project_overrides.pop("bootstrap", None):
            reverse_lookup = {v: k for k, v in bootstrap_options.items()}
            if self.validate_selection_override(bootstraps.keys(), bootstrap_override):
                bootstrap_override = reverse_lookup[bootstraps[bootstrap_override]]

        selected_bootstrap = self.select_option(
            intro=(
                "What GUI toolkit do you want to use for this project?\n"
                "\n"
                "Additional GUI bootstraps are available; visit "
                "https://beeware.org/bee/briefcase-bootstraps "
                "for a full list of known GUI bootstraps."
            ),
            variable="GUI Framework",
            default=None,
            options=bootstrap_options.keys(),
            override_value=bootstrap_override,
        )
        bootstrap_class = bootstrap_options[selected_bootstrap]

        gui_context = {}

        if bootstrap_class is not None:
            bootstrap: BaseGuiBootstrap = bootstrap_class(context=context)

            # Iterate over the Bootstrap interface to build the context.
            # Returning ``None`` is a special case that means the field should not be
            # included in the context and instead deferred to the template default.

            if hasattr(bootstrap, "extra_context"):
                if (additional_context := bootstrap.extra_context()) is not None:
                    gui_context.update(additional_context)

            for context_field in bootstrap.fields:
                if (context_value := getattr(bootstrap, context_field)()) is not None:
                    gui_context[context_field] = context_value

        return gui_context

    def _gui_bootstrap_choices(self, bootstraps):
        """Construct the list of available GUI bootstraps to display to the user."""
        # Sort the options alphabetically first
        ordered = OrderedDict(sorted(bootstraps.items()))

        # Ensure the first 5 options are: Toga, PySide6, Pygame
        ordered.move_to_end("Pygame", last=False)
        ordered.move_to_end("PySide6", last=False)
        ordered.move_to_end("Toga", last=False)

        # Option None should always be last
        ordered["None"] = None
        ordered.move_to_end("None")

        # Construct the bootstrap options as they should be presented to users.
        # The name of the bootstrap is its registered entry point name. Along with the
        # bootstrap's name, a short message important to a user's choice can be shown
        # also; for instance, several show "does not support iOS/Android deployment".
        bootstrap_choices = {}
        max_len = max(map(len, ordered))
        for name, klass in ordered.items():
            if annotation := getattr(klass, "display_name_annotation", ""):
                annotation = f"{' ' * (max_len - len(name))} ({annotation})"
            bootstrap_choices[f"{name}{annotation or ''}"] = klass

        return bootstrap_choices

    def warn_unused_overrides(self, project_overrides: dict[str, str] | None):
        """Inform user of project configuration overrides that were not used."""
        if project_overrides:
            unused_overrides = "\n    ".join(
                f"{key} = {value}" for key, value in project_overrides.items()
            )
            self.logger.warning()
            self.logger.warning(
                "WARNING: These project configuration overrides were not used:\n\n"
                f"    {unused_overrides}"
            )

    def new_app(
        self,
        template: str | None = None,
        template_branch: str | None = None,
        project_overrides: dict[str, str] | None = None,
        **options,
    ):
        """Ask questions to generate a new application, and generate a stub project from
        the briefcase-template."""
        self.input.prompt()
        self.input.prompt("Let's build a new Briefcase app!")

        context = self.build_context(project_overrides=project_overrides)
        self.prompt_divider()  # close the prompting section of output

        self.warn_unused_overrides(project_overrides)

        self.logger.info(
            f"Generating a new application {context['formal_name']!r}",
            prefix=context["app_name"],
        )

        # Make extra sure we won't clobber an existing application.
        if (self.base_path / context["app_name"]).exists():
            raise BriefcaseCommandError(
                f"A directory named {context['app_name']!r} already exists."
            )

        # Create the project files
        self.generate_template(
            template=(
                template
                if template
                else "https://github.com/beeware/briefcase-template"
            ),
            branch=template_branch,
            output_path=self.base_path,
            extra_context=context,
        )

        self.logger.info(
            f"Generated new application {context['formal_name']!r}",
            prefix=context["app_name"],
        )
        self.logger.info(
            f"""
To run your application, type:

    $ cd {context['app_name']}
    $ briefcase dev

"""
        )

    def verify_tools(self):
        """Verify that the tools needed to run this command exist.

        Raises MissingToolException if a required system tool is missing.
        """
        super().verify_tools()
        Git.verify(tools=self.tools)

    def __call__(
        self,
        template: str | None = None,
        template_branch: str | None = None,
        project_overrides: list[str] = None,
        **options,
    ):
        # Confirm host compatibility, and that all required tools are available.
        # There are no apps, so finalize() will be a no op on app configurations.
        self.finalize()

        return self.new_app(
            template=template,
            template_branch=template_branch,
            project_overrides=parse_project_overrides(project_overrides),
            **options,
        )



from __future__ import annotations

from abc import abstractmethod

from briefcase.config import AppConfig

from .base import BaseCommand, full_options


class PackageCommand(BaseCommand):
    command = "package"
    description = "Package an app for distribution."

    ADHOC_SIGN_HELP = "Ignored; signing is not supported"
    IDENTITY_HELP = "Ignored; signing is not supported"

    @property
    def packaging_formats(self):
        return [self.output_format]

    @property
    def default_packaging_format(self):
        return self.output_format

    @abstractmethod
    def distribution_path(self, app):
        """The path to the distributable artefact for the app.

        Requires that the packaging format has been annotated onto the application
        definition

        This is the single file that should be uploaded for distribution. This may be
        the binary (if the binary is a self-contained executable); however, if the
        output format produces an installer, it will be the path to the installer.

        :param app: The app config
        """

    def package_app(self, app: AppConfig, **options):
        """Package an application.

        :param app: The application to package
        """
        # Default implementation; nothing to do.

    def _package_app(
        self,
        app: AppConfig,
        update: bool,
        packaging_format: str,
        **options,
    ) -> dict | None:
        """Internal method to invoke packaging on a single app. Ensures the app exists,
        and has been updated (if requested) before attempting to issue the actual
        package command.

        :param app: The application to package
        :param update: Should the application be updated (and rebuilt) first?
        :param packaging_format: The format of the packaging artefact to create.
        """

        template_file = self.bundle_path(app)
        binary_file = self.binary_path(app)
        if not template_file.exists():
            state = self.create_command(app, **options)
            state = self.build_command(app, **full_options(state, options))
        elif update:
            # If we're updating for packaging, update everything.
            # This ensures everything in the packaged artefact is up to date,
            # and is in a production state
            state = self.update_command(
                app,
                update_resources=True,
                update_requirements=True,
                update_support=True,
                **options,
            )
            state = self.build_command(app, **full_options(state, options))
        elif not binary_file.exists():
            state = self.build_command(app, **options)
        else:
            state = None

        # Annotate the packaging format onto the app
        app.packaging_format = packaging_format

        # Verify the app, which will do final confirmation that we can
        # package in the requested format.
        self.verify_app(app)

        # If the distribution artefact already exists, remove it.
        if self.distribution_path(app).exists():
            self.distribution_path(app).unlink()
        else:
            # Ensure the dist folder exists.
            self.dist_path.mkdir(exist_ok=True)

        # Package the app
        state = self.package_app(app, **full_options(state, options))

        filename = self.distribution_path(app).relative_to(self.base_path)
        self.logger.info(f"Packaged {filename}", prefix=app.app_name)
        return state

    def add_options(self, parser):
        parser.add_argument(
            "-u",
            "--update",
            action="store_true",
            help="Update the app before building",
        )
        parser.add_argument(
            "-p",
            "--packaging-format",
            dest="packaging_format",
            help="Packaging format to use",
            default=self.default_packaging_format,
            choices=self.packaging_formats,
        )

        # --adhoc-sign and --identity are mutually exclusive
        signing_group = parser.add_mutually_exclusive_group()
        signing_group.add_argument(
            "--adhoc-sign",
            help=self.ADHOC_SIGN_HELP,
            action="store_true",
        )
        signing_group.add_argument(
            "-i",
            "--identity",
            dest="identity",
            help=self.IDENTITY_HELP,
            required=False,
        )

    def __call__(
        self,
        app: AppConfig | None = None,
        update: bool = False,
        **options,
    ) -> dict | None:
        # Confirm host compatibility, that all required tools are available,
        # and that the app configuration is finalized.
        self.finalize(app)

        if app:
            state = self._package_app(app, update=update, **options)
        else:
            state = None
            for app_name, app in sorted(self.apps.items()):
                state = self._package_app(
                    app, update=update, **full_options(state, options)
                )

        return state



import pytest

from briefcase.commands.create import Git
from briefcase.exceptions import BriefcaseCommandError


def test_no_git(tracking_create_command, monkeypatch):
    """If Git is not installed, an error is raised."""

    def monkeypatch_verify_git(*a, **kw):
        raise BriefcaseCommandError("Briefcase requires git, but it is not installed")

    monkeypatch.setattr(Git, "verify", monkeypatch_verify_git)

    # The command will fail tool verification.
    with pytest.raises(
        BriefcaseCommandError,
        match=r"Briefcase requires git, but it is not installed",
    ):
        tracking_create_command()


def test_create(tracking_create_command, tmp_path):
    """The create command can be called."""
    tracking_create_command()

    # The right sequence of things will be done
    assert tracking_create_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
        # App configs have been finalized
        ("finalize-app-config", "first"),
        ("finalize-app-config", "second"),
        # Create the first app
        ("generate", "first"),
        ("support", "first"),
        ("verify-app-template", "first"),
        ("verify-app-tools", "first"),
        ("code", "first", False),
        ("requirements", "first", False),
        ("resources", "first"),
        ("cleanup", "first"),
        # Create the second app
        ("generate", "second"),
        ("support", "second"),
        ("verify-app-template", "second"),
        ("verify-app-tools", "second"),
        ("code", "second", False),
        ("requirements", "second", False),
        ("resources", "second"),
        ("cleanup", "second"),
    ]

    # New app content has been created
    assert (tmp_path / "base_path/build/first/tester/dummy/new").exists()
    assert (tmp_path / "base_path/build/second/tester/dummy/new").exists()


def test_create_single(tracking_create_command, tmp_path):
    """The create command can be called to create a single app from the config."""
    tracking_create_command(app=tracking_create_command.apps["first"])

    # The right sequence of things will be done
    assert tracking_create_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
        # App config has been finalized
        ("finalize-app-config", "first"),
        # Create the first app
        ("generate", "first"),
        ("support", "first"),
        ("verify-app-template", "first"),
        ("verify-app-tools", "first"),
        ("code", "first", False),
        ("requirements", "first", False),
        ("resources", "first"),
        ("cleanup", "first"),
    ]

    # New app content has been created
    assert (tmp_path / "base_path/build/first/tester/dummy/new").exists()
    assert not (tmp_path / "base_path/build/second/tester/dummy/new").exists()



# The run command inherits most of its behavior from the common base
# implementation. Do a surface-level verification here, but the app
# tests provide the actual test coverage.
import subprocess
from unittest import mock

import pytest

from briefcase.console import Console, Log
from briefcase.integrations.subprocess import Subprocess
from briefcase.platforms.windows.visualstudio import WindowsVisualStudioRunCommand


@pytest.fixture
def run_command(tmp_path):
    command = WindowsVisualStudioRunCommand(
        logger=Log(),
        console=Console(),
        base_path=tmp_path / "base_path",
        data_path=tmp_path / "briefcase",
    )
    command.tools.home_path = tmp_path / "home"
    command.tools.subprocess = mock.MagicMock(spec_set=Subprocess)

    command._stream_app_logs = mock.MagicMock()

    return command


def test_run_app(run_command, first_app_config, tmp_path):
    """A windows Visual Studio project app can be started."""

    # Set up the log streamer to return a known stream with a good returncode
    log_popen = mock.MagicMock()
    run_command.tools.subprocess.Popen.return_value = log_popen

    # Run the app
    run_command.run_app(first_app_config, test_mode=False, passthrough=[])

    # Popen was called
    run_command.tools.subprocess.Popen.assert_called_with(
        [
            tmp_path
            / "base_path/build/first-app/windows/visualstudio/x64/Release/First App.exe"
        ],
        cwd=tmp_path / "home",
        encoding="UTF-8",
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        bufsize=1,
    )

    # The streamer was started
    run_command._stream_app_logs.assert_called_once_with(
        first_app_config,
        popen=log_popen,
        test_mode=False,
        clean_output=False,
    )


def test_run_app_with_args(run_command, first_app_config, tmp_path):
    """A windows Visual Studio project app can be started with args."""

    # Set up the log streamer to return a known stream with a good returncode
    log_popen = mock.MagicMock()
    run_command.tools.subprocess.Popen.return_value = log_popen

    # Run the app with args
    run_command.run_app(
        first_app_config,
        test_mode=False,
        passthrough=["foo", "--bar"],
    )

    # Popen was called
    run_command.tools.subprocess.Popen.assert_called_with(
        [
            tmp_path
            / "base_path/build/first-app/windows/visualstudio/x64/Release/First App.exe",
            "foo",
            "--bar",
        ],
        cwd=tmp_path / "home",
        encoding="UTF-8",
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        bufsize=1,
    )

    # The streamer was started
    run_command._stream_app_logs.assert_called_once_with(
        first_app_config,
        popen=log_popen,
        test_mode=False,
        clean_output=False,
    )


def test_run_app_test_mode(run_command, first_app_config, tmp_path):
    """A windows Visual Studio project app can be started in test mode."""

    # Set up the log streamer to return a known stream with a good returncode
    log_popen = mock.MagicMock()
    run_command.tools.subprocess.Popen.return_value = log_popen

    # Run the app in test mode
    run_command.run_app(first_app_config, test_mode=True, passthrough=[])

    # Popen was called
    run_command.tools.subprocess.Popen.assert_called_with(
        [
            tmp_path
            / "base_path/build/first-app/windows/visualstudio/x64/Release/First App.exe"
        ],
        cwd=tmp_path / "home",
        encoding="UTF-8",
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        bufsize=1,
        env={"BRIEFCASE_MAIN_MODULE": "tests.first_app"},
    )

    # The streamer was started
    run_command._stream_app_logs.assert_called_once_with(
        first_app_config,
        popen=log_popen,
        test_mode=True,
        clean_output=False,
    )


def test_run_app_test_mode_with_args(run_command, first_app_config, tmp_path):
    """A windows Visual Studio project app can be started in test mode with args."""

    # Set up the log streamer to return a known stream with a good returncode
    log_popen = mock.MagicMock()
    run_command.tools.subprocess.Popen.return_value = log_popen

    # Run the app with args
    run_command.run_app(
        first_app_config,
        test_mode=True,
        passthrough=["foo", "--bar"],
    )

    # Popen was called
    run_command.tools.subprocess.Popen.assert_called_with(
        [
            tmp_path
            / "base_path/build/first-app/windows/visualstudio/x64/Release/First App.exe",
            "foo",
            "--bar",
        ],
        cwd=tmp_path / "home",
        encoding="UTF-8",
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        bufsize=1,
        env={"BRIEFCASE_MAIN_MODULE": "tests.first_app"},
    )

    # The streamer was started
    run_command._stream_app_logs.assert_called_once_with(
        first_app_config,
        popen=log_popen,
        test_mode=True,
        clean_output=False,
    )



import sys
from unittest.mock import MagicMock

import pytest

from briefcase.exceptions import BriefcaseCommandError
from briefcase.integrations.docker import DockerAppContext


def test_match(create_command, first_app_config, capsys):
    """If the system python matches the target python, everything is OK."""
    first_app_config.python_version_tag = "3"
    first_app_config.target_image = "somevendor:surprising"

    create_command.tools[first_app_config].app_context = DockerAppContext(
        tools=create_command.tools,
        app=first_app_config,
    )

    # Mock a return value from Docker that matches the system Python
    create_command.tools[first_app_config].app_context.check_output = MagicMock(
        return_value=f"3.{sys.version_info.minor}\n"
    )

    # Verify python for the app
    create_command.verify_python(first_app_config)

    # The docker container was interrogated for a Python version
    create_command.tools[
        first_app_config
    ].app_context.check_output.assert_called_once_with(
        [
            "python3",
            "-c",
            (
                "import sys; "
                "print(f'{sys.version_info.major}.{sys.version_info.minor}')"
            ),
        ]
    )

    # Warning message was not recorded
    assert "WARNING: Python version mismatch!" not in capsys.readouterr().out

    # The python version has been updated
    assert first_app_config.python_version_tag == f"3.{sys.version_info.minor}"


def test_mismatch(create_command, first_app_config, capsys):
    """If the system python doesn't match the target python, a warning is raised."""
    first_app_config.python_version_tag = "3"
    first_app_config.target_image = "somevendor:surprising"

    create_command.tools[first_app_config].app_context = DockerAppContext(
        tools=create_command.tools,
        app=first_app_config,
    )

    # Mock a return value from Docker that matches the system Python
    create_command.tools[first_app_config].app_context.check_output = MagicMock(
        return_value="3.42\n"
    )

    # Verify python for the app
    create_command.verify_python(first_app_config)

    # The docker container was interrogated for a Python version
    create_command.tools[
        first_app_config
    ].app_context.check_output.assert_called_once_with(
        [
            "python3",
            "-c",
            (
                "import sys; "
                "print(f'{sys.version_info.major}.{sys.version_info.minor}')"
            ),
        ]
    )

    # Warning message was recorded
    assert "WARNING: Python version mismatch!" in capsys.readouterr().out

    # The python version has been updated
    assert first_app_config.python_version_tag == "3.42"


def test_target_too_old(create_command, first_app_config):
    """If the target python is too old, an error is raised."""
    first_app_config.python_version_tag = "3"
    first_app_config.target_image = "somevendor:surprising"

    create_command.logger.warning = MagicMock()
    create_command.tools[first_app_config].app_context = DockerAppContext(
        tools=create_command.tools,
        app=first_app_config,
    )

    # Mock a return value from Docker that is too old for Briefcase
    create_command.tools[first_app_config].app_context.check_output = MagicMock(
        return_value="3.7.16\n"
    )

    # Verify python for the app
    with pytest.raises(
        BriefcaseCommandError,
        match=r"The system python3 version provided by somevendor:surprising "
        r"is 3\.7\.16; Briefcase requires a minimum Python3 version of 3\.8\.",
    ):
        create_command.verify_python(first_app_config)

    # The docker container was interrogated for a Python version
    create_command.tools[
        first_app_config
    ].app_context.check_output.assert_called_once_with(
        [
            "python3",
            "-c",
            (
                "import sys; "
                "print(f'{sys.version_info.major}.{sys.version_info.minor}')"
            ),
        ]
    )



def test_open(open_command, first_app, second_app):
    """The open command can be called."""
    # Configure no command line options
    options, _ = open_command.parse_options([])

    open_command(**options)

    # The right sequence of things will be done
    assert open_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
        # App configs have been finalized
        ("finalize-app-config", "first"),
        ("finalize-app-config", "second"),
        # App template is verified
        ("verify-app-template", "first"),
        # App tools are verified
        ("verify-app-tools", "first"),
        # open the first app
        ("open", "first"),
        # App template is verified
        ("verify-app-template", "second"),
        # App tools are verified
        ("verify-app-tools", "second"),
        # open the second app
        ("open", "second"),
    ]


def test_open_single(open_command, first_app):
    """The open command can be called to open a single app from the config."""
    # Configure no command line options
    options, _ = open_command.parse_options([])

    open_command(app=open_command.apps["first"], **options)

    # The right sequence of things will be done
    assert open_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
        # App config has been finalized
        ("finalize-app-config", "first"),
        # App template is verified
        ("verify-app-template", "first"),
        # App tools are verified
        ("verify-app-tools", "first"),
        # open the first app
        ("open", "first"),
    ]


def test_create_before_open(open_command, tmp_path):
    """If the app doesn't exist, it will be created before opening."""
    # Configure no command line options
    options, _ = open_command.parse_options([])

    open_command(app=open_command.apps["first"], **options)

    # The right sequence of things will be done
    assert open_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
        # App config has been finalized
        ("finalize-app-config", "first"),
        # create, then open the first app
        ("create", "first", {}),
        # App template is verified
        ("verify-app-template", "first"),
        # App tools are verified
        ("verify-app-tools", "first"),
        ("open", "first"),
    ]



from __future__ import annotations

from types import ModuleType

from briefcase.exceptions import BriefcaseCommandError
from briefcase.integrations.base import Tool, ToolCache


class Git(Tool):
    name = "git"
    full_name = "Git"

    @classmethod
    def verify_install(cls, tools: ToolCache, **kwargs) -> ModuleType:
        """Verify if git is installed.

        Unfortunately, `import git` triggers a call on the operating system
        to run the git executable. On some platforms (notably macOS), the git
        binary has been instrumented such that if git *isn't* installed,
        running git triggers a prompt to install Xcode. However, that messes
        with the UX workflow.

        So - we defer importing git until we actually know we need it. This
        enables Briefcase to start to do other Xcode checks as part of the macOS
        workflows, and ensures that "briefcase --help" works on other platforms
        without raising an error.

        :param tools: ToolCache of available tools
        :returns: The git module, if `git` is installed and available.
        """
        # short circuit since already verified and available
        if hasattr(tools, "git"):
            return tools.git

        # Check whether the git executable could be imported.
        try:
            import git
        except ImportError as e:  # pragma: no cover
            # macOS provides git as part of the Xcode command line tools,
            # and also hijacks /usr/bin/git with a trigger that prompts the
            # installation of those tools. Customize the message to account
            # for this.
            if tools.host_os == "Darwin":
                raise BriefcaseCommandError(
                    """\
Briefcase requires git, but it is not installed. Xcode provides git; you should
be shown a dialog prompting you to install Xcode and the Command Line Developer
Tools. Select "Install" to install the Command Line Developer Tools.

Alternatively, you can visit:

    https://git-scm.com/

to download and install git manually.

If you have installed git recently and are still getting this error, you may
need to restart your terminal session.
"""
                ) from e

            else:
                raise BriefcaseCommandError(
                    """\
Briefcase requires git, but it is not installed (or is not on your PATH). Visit:

    https://git-scm.com/

to download and install git manually.

If you have installed git recently and are still getting this error, you may
need to restart your terminal session.
"""
                ) from e

        tools.logger.configure_stdlib_logging("git")

        tools.git = git
        return git



import sys

import pytest

from briefcase.console import Console, Log
from briefcase.exceptions import UnsupportedHostError
from briefcase.platforms.windows.app import WindowsAppCreateCommand


@pytest.fixture
def create_command(tmp_path):
    return WindowsAppCreateCommand(
        logger=Log(),
        console=Console(),
        base_path=tmp_path / "base_path",
        data_path=tmp_path / "briefcase",
    )


@pytest.mark.parametrize("host_os", ["Darwin", "Linux", "WeirdOS"])
def test_unsupported_host_os(create_command, host_os):
    """Error raised for an unsupported OS."""
    create_command.tools.host_os = host_os

    with pytest.raises(
        UnsupportedHostError,
        match="Windows applications can only be built on Windows.",
    ):
        create_command()


@pytest.mark.parametrize("host_arch", ["i686", "ARM64", "wonky"])
def test_unsupported_arch(create_command, host_arch):
    """Windows commands can only run on x86-64."""
    create_command.tools.host_os = "Windows"
    create_command.tools.host_arch = host_arch

    with pytest.raises(
        UnsupportedHostError,
        match=f"Windows applications cannot be built on an {host_arch} machine.",
    ):
        create_command()


def test_supported_arch(create_command):
    """Windows command are allowed to run on x86-64."""
    create_command.tools.host_os = "Windows"
    create_command.tools.host_arch = "AMD64"

    create_command()


def test_unsupported_32bit_python(create_command):
    """Windows commands cannot run with 32bit Python."""
    create_command.tools.host_os = "Windows"
    create_command.tools.host_arch = "AMD64"
    create_command.tools.is_32bit_python = True

    with pytest.raises(
        UnsupportedHostError,
        match="Windows applications cannot be built using a 32bit version of Python",
    ):
        create_command()


@pytest.mark.parametrize(
    "version, version_triple",
    [
        ("1", "1.0.0"),
        ("1.2", "1.2.0"),
        ("1.2.3", "1.2.3"),
        ("1.2.3.4", "1.2.3"),
        ("1.2.3a4", "1.2.3"),
        ("1.2.3b5", "1.2.3"),
        ("1.2.3rc6", "1.2.3"),
        ("1.2.3.dev7", "1.2.3"),
        ("1.2.3.post8", "1.2.3"),
    ],
)
def test_version_triple(
    create_command, first_app_config, tmp_path, version, version_triple
):
    first_app_config.version = version
    context = create_command.output_format_template_context(first_app_config)

    assert context["version_triple"] == version_triple


def test_explicit_version_triple(create_command, first_app_config, tmp_path):
    first_app_config.version = "1.2.3a1"
    first_app_config.version_triple = "2.3.4"

    context = create_command.output_format_template_context(first_app_config)

    # Explicit version triple is used.
    assert context["version_triple"] == "2.3.4"


def test_guid(create_command, first_app_config, tmp_path):
    """A predictable GUID will be generated from the bundle."""
    context = create_command.output_format_template_context(first_app_config)

    assert context["guid"] == "d666a4f1-c7b7-52cc-888a-3a35a7cc97e5"


def test_explicit_guid(create_command, first_app_config, tmp_path):
    """If a GUID is explicitly provided, it is used."""
    first_app_config.guid = "e822176f-b755-589f-849c-6c6600f7efb1"
    context = create_command.output_format_template_context(first_app_config)

    # Explicitly provided GUID is used.
    assert context["guid"] == "e822176f-b755-589f-849c-6c6600f7efb1"


def test_support_package_url(create_command, first_app_config, tmp_path):
    """A valid support package URL is created for a support revision."""
    revision = 5
    expected_link = (
        f"https://www.python.org/ftp/python"
        f"/{sys.version_info.major}.{sys.version_info.minor}.{revision}"
        f"/python-{sys.version_info.major}.{sys.version_info.minor}.{revision}-embed-amd64.zip"
    )
    assert create_command.support_package_url(revision) == expected_link


def test_default_install_scope(create_command, first_app_config, tmp_path):
    """By default, app should be installed per user."""
    context = create_command.output_format_template_context(first_app_config)

    assert context == {
        "guid": "d666a4f1-c7b7-52cc-888a-3a35a7cc97e5",
        "version_triple": "0.0.1",
        "install_scope": None,
    }


def test_per_machine_install_scope(create_command, first_app_config, tmp_path):
    """By default, app should be installed per user."""
    first_app_config.system_installer = True

    context = create_command.output_format_template_context(first_app_config)

    assert context == {
        "guid": "d666a4f1-c7b7-52cc-888a-3a35a7cc97e5",
        "version_triple": "0.0.1",
        "install_scope": "perMachine",
    }


def test_per_user_install_scope(create_command, first_app_config, tmp_path):
    """App can be set to have explicit per-user scope."""
    first_app_config.system_installer = False

    context = create_command.output_format_template_context(first_app_config)

    assert context == {
        "guid": "d666a4f1-c7b7-52cc-888a-3a35a7cc97e5",
        "version_triple": "0.0.1",
        "install_scope": "perUser",
    }



from unittest import mock

import pytest

from briefcase.commands import DevCommand
from briefcase.config import AppConfig
from briefcase.console import Console, Log
from briefcase.integrations.subprocess import Subprocess


@pytest.fixture
def dev_command(tmp_path):
    command = DevCommand(logger=Log(), console=Console(), base_path=tmp_path)
    command.tools.subprocess = mock.MagicMock(spec_set=Subprocess)
    return command


@pytest.fixture
def first_app_uninstalled(tmp_path):
    # Make sure the source code exists
    (tmp_path / "src/first").mkdir(parents=True, exist_ok=True)
    with (tmp_path / "src/first/__init__.py").open("w", encoding="UTF-8") as f:
        f.write('print("Hello world")')

    return AppConfig(
        app_name="first",
        bundle="com.example",
        version="0.0.1",
        description="The first simple app",
        sources=["src/first"],
        license={"file": "LICENSE"},
    )


@pytest.fixture
def first_app(tmp_path, first_app_uninstalled):
    # The same fixture as first_app_uninstalled,
    # but ensures that the .dist-info folder for the app exists
    (tmp_path / "src/first.dist-info").mkdir(exist_ok=True)
    return first_app_uninstalled


@pytest.fixture
def second_app(tmp_path):
    # Make sure the source code exists
    (tmp_path / "src/second").mkdir(parents=True, exist_ok=True)
    with (tmp_path / "src/second/__init__.py").open("w", encoding="UTF-8") as f:
        f.write('print("Hello world")')

    # Create the dist-info folder
    (tmp_path / "src/second.dist-info").mkdir(exist_ok=True)

    return AppConfig(
        app_name="second",
        bundle="com.example",
        version="0.0.2",
        description="The second simple app",
        sources=["src/second"],
        license={"file": "LICENSE"},
    )


@pytest.fixture
def third_app(tmp_path):
    # Make sure the source code exists
    (tmp_path / "src/third").mkdir(parents=True, exist_ok=True)
    with (tmp_path / "src/third/__init__.py").open("w", encoding="UTF-8") as f:
        f.write('print("Hello world")')

    # Create the dist-info folder
    (tmp_path / "src/third.dist-info").mkdir(exist_ok=True)

    return AppConfig(
        app_name="third",
        bundle="com.example",
        version="0.0.2",
        description="The third simple app",
        sources=["src/third", "src/common", "other"],
        test_sources=["tests", "path/to/other"],
        license={"file": "LICENSE"},
    )



import subprocess
from unittest.mock import Mock


def test_pid_exists_succeed(adb):
    """adb.pid_exists() can be called on a process that exists."""
    adb.run = Mock(return_value="")

    assert adb.pid_exists("1234")
    adb.run.assert_called_once_with("shell", "test", "-e", "/proc/1234")


def test_pid_exists_quiet(adb):
    """adb.pid_exists() can be called in quiet mode on a process that exists."""
    adb.run = Mock(return_value="")

    assert adb.pid_exists("1234", quiet=True)
    adb.run.assert_called_once_with("shell", "test", "-e", "/proc/1234", quiet=True)


def test_pid_does_not_exist(adb):
    """If adb.pid_exists() returns a status code of 1, it is interpreted as the process
    not existing."""
    adb.run = Mock(side_effect=subprocess.CalledProcessError(returncode=1, cmd="test"))

    assert not adb.pid_exists("9999") is None
    adb.run.assert_called_once_with("shell", "test", "-e", "/proc/9999")



import sys
from pathlib import Path

import pytest

from briefcase.__main__ import main
from briefcase.commands.create import CreateCommand
from briefcase.commands.run import RunCommand
from briefcase.console import Log
from briefcase.exceptions import BriefcaseTestSuiteFailure, BriefcaseWarning

from .utils import create_file


@pytest.fixture
def pyproject_toml(monkeypatch, tmp_path):
    # Monkeypatch cwd() to use a test folder.
    monkeypatch.setattr(Path, "cwd", lambda: tmp_path)

    create_file(
        tmp_path / "pyproject.toml",
        """
[build-system]
requires = ["briefcase"]

[tool.briefcase]
project_name = "Hello World"
bundle = "com.example"
version = "0.0.1"
license.file = "LICENSE"

[tool.briefcase.app.myapp]
description = "My first application"
sources = ["myapp"]

    """,
    )


def test_help(monkeypatch, tmp_path, capsys):
    """Briefcase can output help."""
    # Set the test command line
    monkeypatch.setattr(sys, "argv", ["briefcase", "--help"])

    # Help has a return code of -10
    assert main() == -10

    output = capsys.readouterr().out
    assert output.startswith(
        "\nusage: briefcase [-h] <command> [<platform>] [<format>] ...\n"
    )

    # No log file was written
    assert len(list(tmp_path.glob(f"{Log.LOG_DIR}/briefcase.*.log"))) == 0


def test_command(monkeypatch, tmp_path, capsys):
    """A command can be successful."""
    # Monkeypatch cwd() to use a test folder.
    monkeypatch.setattr(Path, "cwd", lambda: tmp_path)

    # Create a dummy empty template to use in the new command
    create_file(tmp_path / "template/cookiecutter.json", '{"app_name": "app_name"}')
    create_file(tmp_path / "template/{{ cookiecutter.app_name }}/app", "content")

    # Set the test command line
    monkeypatch.setattr(
        sys,
        "argv",
        ["briefcase", "new", "--no-input", "--template", str(tmp_path / "template")],
    )

    # Successful return is 0
    assert main() == 0

    output = capsys.readouterr().out
    assert output.startswith(
        "\n[helloworld] Generating a new application 'Hello World'"
    )

    # No log file was written
    assert len(list(tmp_path.glob(f"{Log.LOG_DIR}/briefcase.*.log"))) == 0


def test_command_warning(monkeypatch, pyproject_toml, tmp_path, capsys):
    """A command can raise a known warning."""
    # Monkeypatch cwd() to use a test folder.
    monkeypatch.setattr(Path, "cwd", lambda: tmp_path)

    # Set the test command line
    monkeypatch.setattr(sys, "argv", ["briefcase", "create"])

    # Monkeypatch a warning into the create command
    def sort_of_bad_generate_app_template(self, app):
        raise BriefcaseWarning(error_code=0, msg="This is bad, but not *really* bad")

    monkeypatch.setattr(
        CreateCommand, "generate_app_template", sort_of_bad_generate_app_template
    )

    # The error code has been returned
    assert main() == 0

    # The warning has been output
    output = capsys.readouterr().out
    assert output.endswith("\nThis is bad, but not *really* bad\n")

    # Log files are not created for BriefcaseWarnings
    assert len(list(tmp_path.glob(f"{Log.LOG_DIR}/briefcase.*.create.log"))) == 0


def test_command_error(monkeypatch, tmp_path, capsys):
    """A command can raise a known error."""
    # Monkeypatch cwd() to use a test folder.
    monkeypatch.setattr(Path, "cwd", lambda: tmp_path)

    # Set the test command line
    monkeypatch.setattr(sys, "argv", ["briefcase", "create"])

    # BriefcaseConfigError has a return code of 100
    assert main() == 100

    output = capsys.readouterr().out
    assert output.startswith(
        "\nBriefcase configuration error: Configuration file not found."
    )

    # Log files are not created for BriefcaseConfigError errors
    assert len(list(tmp_path.glob(f"{Log.LOG_DIR}/briefcase.*.create.log"))) == 0


def test_unknown_command_error(monkeypatch, pyproject_toml, capsys):
    """A command can raise an unknown error."""
    monkeypatch.setattr(sys, "argv", ["briefcase", "create"])

    # Monkeypatch an error into the create command
    def bad_generate_app_template(self, app):
        raise ValueError("Bad value")

    monkeypatch.setattr(
        CreateCommand, "generate_app_template", bad_generate_app_template
    )

    # Error is surfaced to the user
    with pytest.raises(ValueError, match=r"Bad value"):
        main()


def test_interrupted_command(monkeypatch, pyproject_toml, tmp_path, capsys):
    """A command can be interrupted."""
    monkeypatch.setattr(sys, "argv", ["briefcase", "create"])

    # Monkeypatch a keyboard interrupt into the create command
    def interrupted_generate_app_template(self, app):
        raise KeyboardInterrupt()

    monkeypatch.setattr(
        CreateCommand, "generate_app_template", interrupted_generate_app_template
    )

    # Interrupted command return -42
    assert main() == -42

    output = capsys.readouterr().out
    assert "\nAborted by user.\n" in output

    # No log file was written
    assert len(list(tmp_path.glob(f"{Log.LOG_DIR}/briefcase.*.create.log"))) == 0


def test_interrupted_command_with_log(monkeypatch, pyproject_toml, tmp_path, capsys):
    """A log can be generated when a command is interrupted."""
    monkeypatch.setattr(sys, "argv", ["briefcase", "create", "--log"])

    # Monkeypatch a keyboard interrupt into the create command
    def interrupted_generate_app_template(self, app):
        raise KeyboardInterrupt()

    monkeypatch.setattr(
        CreateCommand, "generate_app_template", interrupted_generate_app_template
    )

    # Interrupted command return -42
    assert main() == -42

    output = capsys.readouterr().out
    assert "\nAborted by user.\n" in output

    # A log file was written
    assert len(list(tmp_path.glob(f"{Log.LOG_DIR}/briefcase.*.create.log"))) == 1


def test_test_failure(monkeypatch, pyproject_toml, tmp_path, capsys):
    """A test suite failure can be reported."""
    monkeypatch.setattr(sys, "argv", ["briefcase", "run", "--test"])

    # Monkeypatch a keyboard interrupt into the run command
    def run_call(self, **kwargs):
        raise BriefcaseTestSuiteFailure()

    monkeypatch.setattr(RunCommand, "__call__", run_call)

    # Failed test suite returns 1000
    assert main() == 1000

    # The code generating the test failure is responsible for output
    output = capsys.readouterr().out
    assert output == ""

    # Log files are not created for BriefcaseTestSuiteFailures
    assert len(list(tmp_path.glob(f"{Log.LOG_DIR}/briefcase.*.create.log"))) == 0



from unittest.mock import MagicMock

import pytest

from briefcase.integrations.cookiecutter import UUIDExtension


@pytest.mark.parametrize(
    "value, expected",
    [
        ("example.com", "cfbff0d1-9375-5685-968c-48ce8b15ae17"),
        ("foobar.example.com", "941bbcd9-03e1-568a-a728-8434055bc338"),
    ],
)
def test_dns_uuid5_value(value, expected):
    env = MagicMock()
    env.filters = {}
    UUIDExtension(env)
    assert env.filters["dns_uuid5"](value) == expected



import subprocess
from unittest.mock import MagicMock

import pytest

from briefcase.exceptions import BriefcaseCommandError, InvalidDeviceError


def test_booted(adb, capsys):
    """A booted device returns true."""
    # Mock out the adb response for an emulator
    adb.run = MagicMock(return_value="1\n")

    # Invoke avd_name
    assert adb.has_booted()

    # Validate call parameters.
    adb.run.assert_called_once_with("shell", "getprop", "sys.boot_completed")


def test_not_booted(adb, capsys):
    """A non-booted device returns False."""
    # Mock out the adb response for an emulator
    adb.run = MagicMock(return_value="\n")

    # Invoke avd_name
    assert not adb.has_booted()

    # Validate call parameters.
    adb.run.assert_called_once_with("shell", "getprop", "sys.boot_completed")


def test_adb_failure(adb, capsys):
    """If ADB fails, an error is raised."""
    # Mock out the adb response for an emulator
    adb.run = MagicMock(
        side_effect=subprocess.CalledProcessError(returncode=69, cmd="emu avd name")
    )

    # Invoke avd_name
    with pytest.raises(BriefcaseCommandError):
        adb.has_booted()

    # Validate call parameters.
    adb.run.assert_called_once_with("shell", "getprop", "sys.boot_completed")


def test_invalid_device(adb, capsys):
    """If the device ID is invalid, an error is raised."""
    # Mock out the adb response for an emulator
    adb.run = MagicMock(side_effect=InvalidDeviceError("device", "exampleDevice"))

    # Invoke avd_name
    with pytest.raises(BriefcaseCommandError):
        adb.has_booted()

    # Validate call parameters.
    adb.run.assert_called_once_with("shell", "getprop", "sys.boot_completed")
